/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package at.dms.kjc;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.HashMap;
import java.util.Map;

import at.dms.compiler.CWarning;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;
import at.dms.util.MessageDescription;

/**
 * This class represents a local context during checkBody.
 *
 * It follows the control flow and maintain informations about
 * variable (initialized, used, allocated), exceptions (thrown, catched).
 * It also verify that context is still reachable
 *
 * There is a set of utilities method to access fields, methods and class
 * with the name by clamping the parsing tree.
 *
 * @see CCompilationUnitContext
 * @see CClassContext
 * @see CMethodContext
 * @see CContext
 */
public class CBlockContext
    extends CAbstractBodyContext
{
    /**
     * Inner (local) classes declared in this block.
     */
    private Map innerClasses;

    /**
     * A list of the local variables in this block.
     */
    private List localVars;

    /**
     * Index of parent in local variable tables.
     */
    private final int parentIndex;

    /**
     * Position for next new local.
     */
    private int localsPosition;

    /**
     * Construct a block context, it supports local variable allocation
     * throw statement and return statement.
     *
     * @param parent
     * The parent context, it must be different than null except if called
     * by the top level.
     */
    CBlockContext(CMethodContext parent, int localVars)
    {
        super(parent);

        this.localVars = (localVars == 0 ? null : new ArrayList(localVars));
        this.localsPosition = 0;
        this.parentIndex = 0;
        this.variableInfo = new CVariableInfo(); // $$$
    }

    /**
     * Construct a block context, it supports local variable allocation
     * throw statement and return statement.
     *
     * @param parent
     * The parent context, it must be different than null except if called
     * by the top level.
     */
    public CBlockContext(CBodyContext parent)
    {
        this(parent, 0);
    }

    /**
     * Construct a block context, it supports local variable allocation
     * throw statement and return statement.
     *
     * @param parent
     * The parent context, it must be different than null except if called
     * by the top level.
     */
    public CBlockContext(CBodyContext parent, int predictedVars)
    {
        super(parent);

        this.localVars = new ArrayList(predictedVars == 0 ? 5 : predictedVars);

        CBlockContext parentBlock = parent.getBlockContext();

        this.localsPosition = parentBlock.getMaxLocalsPosition();
        this.parentIndex = parentBlock.getMaxLocalsIndex();
        this.variableInfo = new CVariableInfo(); // $$$
    }

    public CBodyContext createBodyContext()
    {
        return new CBodyContext(this);
    }

    /**
     * Verify everything is okay at the end of this context
     */
    public void close(TokenReference ref)
    {
        // default, no errors
        verifyLocalVarUsed();

        super.close(ref);
    }

    /**
     * check everything correct
     * @exception UnpositionedError this error will be positioned soon
     */
    private void verifyLocalVarUsed()
    {
        if (localVars == null) {
            return;
        }

        Iterator iterator = localVars.iterator();
        while (iterator.hasNext()) {
            JLocalVariable var = (JLocalVariable)iterator.next();
      
            if (!var.isUsed() && !var.getIdent().startsWith("_")) {
                MessageDescription  mesg = null;
 
                switch (var.getDescription()) {
                case JLocalVariable.DES_PARAMETER:
                    mesg = KjcMessages.UNUSED_PARAMETER;
                    break;
                case JLocalVariable.DES_CATCH_PARAMETER:
                    mesg = KjcMessages.UNUSED_CATCH_PARAMETER;
                    break;
                case JLocalVariable.DES_LOCAL_VAR:
                    mesg = KjcMessages.UNUSED_LOCALVAR;
                    break;
                default:
                    continue;
                }
                reportTrouble(new CWarning(var.getTokenReference(),
                                           mesg,
                                           var.getIdent()));

            }
            else if (var.getDescription() == JLocalVariable.DES_LOCAL_VAR
                     && !var.isFinal()) {
                reportTrouble
                    (new CWarning(var.getTokenReference(),
                                  KjcMessages.CONSTANT_VARIABLE_NOT_FINAL,
                                  var.getIdent()));
            }
        }
    }

    public CBodyContext cloneContext()
    {
        throw new RuntimeException();
    }

    /**
     * @param var  the definition of a variable
     * @return all informations we have about this variable
     */
    public int getVariableInfo(int index)
    {
        if (index < (localVars.size() + parentIndex) && index >= parentIndex) {
            return variableInfo.getInfo(index - parentIndex);
        }
        else {
            return ((CBodyContext)parent).getVariableInfo(index);
        }
    }

    /**
     * @param index  The variable position in method array of local vars
     * @param info  The information to add
     *
     * We make it a local copy of this information and at the end of this
     * context we will transfert it to the parent context according to
     * controlFlow.
     */
    public void setVariableInfo(int index, int info)
    {
        if (index < (localVars.size() + parentIndex) && index >= parentIndex) {
            variableInfo.setInfo(index - parentIndex, info);
        }
        else {
            ((CBodyContext)parent).setVariableInfo(index, info);
        }
    }

    /**
     * addLocal variable.
     *
     * @param var  the name of the variable
     * @param initialized is the varaible already initialized
     * @exception UnpositionedError this error will be positioned soon
     */
    public void addVariable(JLocalVariable var)
        throws UnpositionedError
    {
        if (localVars == null) {
            localVars = new ArrayList();
        }

        if (lookupLocalVariable(var.getIdent()) != null) {
            throw new UnpositionedError(KjcMessages.VAR_RENAME,
                                        var.getIdent());
        }

        var.setPosition(localsPosition);
        var.setIndex(localVars.size() + parentIndex);

        localVars.add(var);
        localsPosition += var.getType().getSize();
    }

    /**
     * Shifts the positions of the locals in this block. This is necessary
     * in constructors of local classes, where a variable number of
     * synthetic constructor parameters (the values of the outer local
     * variables) may be inserted after other local have already been
     * assigned a position.
     */
    public void shiftLocalsPositions(int shift)
    {
    }

    /**
     * lookupLocalVariable
     * @param ident  the name of the variable
     * @return a variable from an ident in current context
     */
    public JLocalVariable lookupLocalVariable(String ident)
    {
        if (localVars != null) {
            Iterator iterator = localVars.iterator();
            while (iterator.hasNext()) {
                JLocalVariable var = (JLocalVariable)iterator.next();

                if (var.getIdent().equals(ident)) {
                    return var;
                }
            }
        }

        return parent.lookupLocalVariable(ident);
    }

    /**
     * Dereference a variable that is not used after (exit block)
     */
    public void dereferenceVariable(JLocalVariable var)
    {
        localVars.remove(var);
        localsPosition -= var.getType().getSize();
    }

    /**
     * addThisVariable: make room for this variable.
     */
    public void addThisVariable()
    {
        localsPosition += 1;
    }

    /**
     * addOuterThisVariable: make room for outerThis variable in a constructor.
     */
    public void addOuterThisVariable()
    {
        localsPosition += 1;
    }

    /**
     * The maximum locals position in the local variables of the bytecode
     * for the method in which this is a block.
     */
    public int getMaxLocalsPosition()
    {
        return localsPosition;
    }

    /**
     * The maximum local index for the method in which this is a block.
     */
    public int getMaxLocalsIndex()
    {
        return parentIndex + ((localVars == null) ? 0 : localVars.size());
    }

    public CBlockContext getBlockContext()
    {
        return this;
    }

    /**
     * addLocalClass
     * @param clazz  the clazz to add
     * @exception UnpositionedError this error will be positioned soon
     */
    public void addClass(CClass clazz) throws UnpositionedError
    {
        if (innerClasses == null) {
            innerClasses = new HashMap();
        }
        Object old = innerClasses.put(clazz.getIdent(), clazz);
        if (old != null) {
            throw new UnpositionedError(KjcMessages.CLAZZ_RENAME,
                                        clazz.getIdent());
        }
    }

    /**
     * lookupLocalClass
     * @param ident  the name of the class
     * @return a class from an ident in current context
     */
    public CClass lookupLocalClass(String ident)
    {
        return (innerClasses == null
                ? null : (CClass)innerClasses.get(ident));
    }

    /**
     * lookupClass
     * search for a class with the provided type parameters
     * @param name  method name
     * @return the class if found, null otherwise
     * @exception UnpositionedError this error will be positioned soon
     */
    public CClassType lookupClass(String name) throws UnpositionedError
    {
        CClass clazz = lookupLocalClass(name);

        if (clazz != null) {
            return clazz.getType();
        }
        else {
            return super.lookupClass(name);
        }
    }
}
