/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import java.util.HashSet;

/**
 * This class represents a local context during checkBody
 * It follows the control flow and maintain informations about
 * variable (initialized, used, allocated), exceptions (thrown, catched)
 * It also verify that context is still reachable
 *
 * There is a set of utilities method to access fields, methods and class
 * with the name by clamping the parsing tree
 * @see CContext
 * @see CCompilationUnitContext
 * @see CClassContext
 * @see CMethodContext
 * @see CConstructorContext
 * @see CInitializerContext
 * @see CBodyContext
 * @see CBlockContext
 */
public class CBodyContext extends CAbstractBodyContext
{
    /**
     * Construct a block context, it supports local variable allocation
     * throw statement and return statement
     * This starts a new context in a method
     * @param parent  the parent context, it must be different
     *    than null except if called by the top level
     */
    protected CBodyContext(CAbstractBodyContext parent)
    {
        super(parent);
    }

    /**
     * Construct a block context, it supports local variable allocation
     * throw statement and return statement
     * This copy the parent context
     * @param parent  the parent context, it must be different
     *    than null except if called by the top level
     */
    private CBodyContext(CAbstractBodyContext parent, CBodyContext clone)
    {
        super(parent);

        flowState = clone.flowState;
        variableInfo = (clone.variableInfo == null
                        ? null : (CVariableInfo)clone.variableInfo.clone());
        fieldInfo = (clone.fieldInfo == null
                     ? null : (CVariableInfo)clone.fieldInfo.clone());
    }

    public CClassContext createClassContext(CSourceClass clazz)
    {
        return new CClassContext(this, clazz, null);
    }

    public CBlockContext createBlockContext()
    {
        return new CBlockContext(this);
    }
  
    public CBlockContext createBlockContext(int params)
    {
        return new CBlockContext(this, params);
    }

    public CLoopContext createLoopContext(JLoopStatement self)
    {
        return new CLoopContext(this, self);
    }
  
    public CTryContext createTryContext()
    {
        return new CTryContext(this);
    }

    public CExpressionContext createExpressionContext()
    {
        return new CExpressionContext(this);
    }

    public CBodyContext cloneContext()
    {
        return new CBodyContext((CAbstractBodyContext)parent, this);
    }

    public CBodyContext makeSnapshot()
    {
        CBodyContext copy = cloneContext();

        copy.throwables = new HashSet();
        return copy;
    }

    public void merge(CBodyContext context)
    { 
        // System.err.println("Merging contexts");
        flowState = flowState & context.flowState;

        if (variableInfo != null || context.variableInfo != null) {
            int parentPosition = getBlockContext().getMaxLocalsIndex();
      
            for (int i = 0; i < parentPosition; i++) {
                int info1 = getVariableInfo(i);
                int info2 = context.getVariableInfo(i);
 
                if (info1 + info2 != 0) {
                    /*
                    System.err.println("Merging pos = " + i + ": "
                                      + info1 + " :+: " + info2);
                    */
                    setVariableInfo(i,
                                    CVariableInfo.merge(info1, info2),
                                    true);
                }
            }
        }

        if (fieldInfo != null || context.fieldInfo != null) {
            int parentPosition =
                parent.getClassContext().getCClass().getFieldCount();
      
            for (int i = 0; i < parentPosition; i++) {
                int info1 = getFieldInfo(i);
                int info2 = context.getFieldInfo(i);
 
                if (info1 + info2 != 0) {
                    if (fieldInfo == null) {
                        fieldInfo = new CVariableInfo();
                    }
                    fieldInfo.setInfo(i, CVariableInfo.merge(info1, info2),
                                      true);
                }
            }    
        }
    }

    public void adopt(CBodyContext context)
    {
        // System.err.println("Adopting context");
        flowState = context.flowState;

        int parentPosition = getBlockContext().getMaxLocalsIndex();
      
        for (int i = 0; i < parentPosition; i++) {
            int info = context.getVariableInfo(i);
      
            if (info != 0) {
                setVariableInfo(i, info);
            }
        }

        parentPosition = parent.getClassContext().getCClass().getFieldCount();
    
        for (int i = 0; i < parentPosition; i++) {
            int info = context.getFieldInfo(i);
      
            if (info != 0) {
                setFieldInfo(i, info);
            }    
        }
    }
}
