/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package at.dms.kjc;

import at.dms.compiler.UnpositionedError;

/**
 * This class represents class type in the type structure.
 */
public class CClassNameType extends CClassType
{
    /**
     * null => checked
     */
    private String qualifiedName;

    /**
     * Construct a class type.
     *
     * @param qualifiedName
     * The class qualified name of the class.
     */
    public CClassNameType(String qualifiedName)
    {
        super();

        // ???: if the name contains a dot, it's bad. So does that mean that
        // it should contain slashes or no qualification at all?
        if (qualifiedName.indexOf('.') >= 0) {
            throw new RuntimeException(); // $$$
        }

        this.qualifiedName = qualifiedName.intern();
    }

    public void setQualifiedName(String qname)
    {
        /* DBG:
        if (qname == null) {
            System.err.println("Old qname = " + this.qualifiedName);
            Thread.dumpStack();
        }
        */
        this.qualifiedName = qname;
    }

    /**
     * Transforms this type to a string
     */
    public String toString()
    {
        return (qualifiedName != null
                ? qualifiedName.replace('/', '.')
                : super.toString());
    }

    /**
     *
     */
    public String getQualifiedName()
    {
        return (qualifiedName == null
                ? super.getQualifiedName()
                : qualifiedName);
    }

    /**
     * Returns the class object associated with this type
     *
     * If this type was never checked (read from class files)
     * check it!
     *
     * @return the class object associated with this type
     */
    public CClass getCClass()
    {
        if (qualifiedName != null) {
            try {
                checkType(null);
            }
            catch (UnpositionedError cue) {
                setClass(new CBadClass(qualifiedName));
                setQualifiedName(null);
                return getCClass();
            }
        }

        return super.getCClass();
    }

    /**
     * Try to resolve an inner type name!
     */
    private boolean tryResolveInnerType(CContext context)
        throws UnpositionedError
    {
        // May be a innerclass
        int index = qualifiedName.lastIndexOf("/");
        if (index > 0) {
            try {
                CClassNameType outer =
                    new CClassNameType(qualifiedName.substring(0, index)
                                       .intern());
                /* DBG:
                System.err.print("Checking outer "
                                 + qualifiedName.substring(0, index)
                                 + "...");
                */
                outer.checkType(context);
                // DBG: System.err.println(qualifiedName.substring(0, index)
                // DBG:                    + " succeeded!");
                String innerName = 
                    qualifiedName.substring(index + 1).intern();

                CClass innerClazz =
                    outer.getCClass().lookupMemberClass(innerName);

                if (innerClazz != null) {
                    setClass(innerClazz);
                    setQualifiedName(null);
                    return true;
                }
                else {
                    return false;
                }
            }
            catch (UnpositionedError ce) {
                /* DBG:
                if (qualifiedName == null) {
                    ce.printStackTrace();
                }
                System.err.println("Failed: " + ce +
                                   (qualifiedName == null
                                    ? " [qname == null]" : ""));
                */
                return false;
            }
        }
        else {
            return false;
        }
    }

    /**
     * Check that type is valid.
     * Necessary to resolve String into java/lang/String.
     *
     * @param context
     * The context (may be be null.)
     *
     * @exception UnpositionedError
     * This error will be positioned soon.
     */
    public void checkType(CContext context)
        throws UnpositionedError
    {
        if (qualifiedName == null) {
            /* Class has already been checked and (hopefully) resolved. */
            return;
        }
        //DBG: System.err.println("Checking type " + qualifiedName);

        String saved = qualifiedName;
        /* First look it up as a member class */
        if (tryResolveInnerType(context)) {
            return;
        }

        if (context != null) {
            /* DBG:
            if (qualifiedName == null) {
                System.err.println("Help, qualifiedName is null and type is "
                                   + super.getCClass().getIdent() +
                                   " and orig qname is " + saved);
            }
            */
            CClassType type = context.lookupClass(qualifiedName);
            if (type != null) {
                if ((type instanceof CClassNameType) && (context != null)
                    && (((CClassNameType)type).qualifiedName != null)) {
                    
                    setQualifiedName(((CClassNameType)type).qualifiedName);
                    // we have found the good one but we dont want to
                    // really load the class here.
                    //setClass(type.getCClass());
                    // setQualifiedName(null);
                }
                else {
                    setClass(type.getCClass());
                    setQualifiedName(null);
                }
                return;
            }
        }

        /* Try to look it up at top level */
        if (CTopLevel.hasClassFile(qualifiedName)) {
            setClassFromName(qualifiedName);
            setQualifiedName(null);
            return;
        }
        // DBG: System.err.println("Not found: " + qualifiedName);
        throw new UnpositionedError(KjcMessages.TYPE_UNKNOWN, qualifiedName);
    }

}
