/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package at.dms.kjc;

import at.dms.compiler.Compiler;
import at.dms.compiler.PositionedError;
import at.dms.compiler.UnpositionedError;
import at.dms.util.MessageDescription;

/**
 * This class represents a local context during checkBody
 * It follows the control flow and maintain informations about
 * variable (initialized, used, allocated), exceptions (thrown, catched)
 * It also verify that context is still reachable
 *
 * There is a set of utilities method to access fields, methods and class
 * with the name by clamping the parsing tree
 * @see CContext
 * @see CCompilationUnitContext
 * @see CClassContext
 * @see CMethodContext
 * @see CBodyContext
 * @see CBlockContext
 */
public abstract class CContext
    extends at.dms.util.Utils
    implements Constants
{

    /**
     * Construct a block context, it supports local variable allocation
     * throw statement and return statement
     * @param parent  the parent context, it must be different
     *    than null except if called by the top level
     */
    public CContext()
    {
        cunit = (CCompilationUnitContext)this;
    }

    /**
     * Construct a block context, it supports local variable allocation
     * throw statement and return statement
     * @param parent  the parent context, it must be different
     *    than null except if called by the top level
     */
    protected CContext(CContext parent)
    {
        this.parent = parent;
        cunit = parent.cunit;
    }

    /**
     * lookupClass
     * search for a class with the provided type parameters
     * @param name  method name
     * @return the class if found, null otherwise
     * @exception UnpositionedError this error will be positioned soon
     */
    public CClassType lookupClass(String name)
        throws UnpositionedError
    {
        return parent.lookupClass(name);
    }

    /**
     * lookupMethod
     * search for a matching method with the provided type parameters
     * @param ident  method name
     * @param params  method parameters
     * @return the method if found, null otherwise
     * @exception UnpositionedError this error will be positioned soon
     */
    public CMethod lookupMethod(CClass caller, String ident, CType[] params)
        throws UnpositionedError
    {
        return parent.getClassContext().lookupMethod(caller, ident, params);
    }

    /**
     * lookupLocalVariable
     * @return a variable from an ident in current context
     * @exception UnpositionedError this error will be positioned soon
     */
    public CField lookupField(String ident) throws UnpositionedError
    {
        return getClassContext().lookupField(ident);
    }

    /**
     * lookupLocalVariable
     * @param ident  the name of the local variable
     * @return a variable from an ident in current context
     */
    public JLocalVariable lookupLocalVariable(String ident)
    {
        return parent.lookupLocalVariable(ident);
    }

    /**
     * @param field  the definition of a field
     * @return a field from a field definition in current context
     * @exception UnpositionedError this error will be positioned soon
     */
    public int getFieldInfo(int index)
    {
        return parent.getFieldInfo(index);
    }

    /**
     * @param field  the definition of a field
     * @return a field from a field definition in current context
     * @exception UnpositionedError this error will be positioned soon
     */
    public void setFieldInfo(int index, int info)
    {
        parent.setFieldInfo(index, info);
    }

    /**
     * getParentContext
     * @return the parent
     */
    public CContext getParentContext()
    {
        return parent;
    }

    /**
     * @return the compilation unit
     */
    public CCompilationUnitContext getCompilationUnit()
    {
        return cunit;
    }

    /**
     * getClassContext
     * @return the near parent of type CClassContext
     */
    public CClassContext getClassContext()
    {
        return parent.getClassContext();
    }

    /**
     * getMethod
     * @return the near parent of type CClassContext
     */
    public CMethodContext getMethodContext()
    {
        return parent.getMethodContext();
    }

    /**
     * Returns the nearest block context (Where yuo can define some local vars)
     */
    public CBlockContext getBlockContext()
    {
        return parent.getBlockContext();
    }

    /**
     * Adds a class to generate
     */
    public void addSourceClass(CSourceClass clazz)
    {
        getCompilationUnit().addSourceClass(clazz);
    }

    /**
     * Add an error into the list and eat it.
     * This method should be called after a try catch block after catching
     * exception or directly without exception thrown.
     * @param error  the error
     */
    public void reportTrouble(PositionedError trouble)
    {
        getCompilationUnit().reportTrouble(trouble);
    }

    /**
     * Add an error into the list and eat it.
     * This method should be called after a try catch block after catching
     * exception or directly without exception thrown.
     * @param mess  the error message
     * @param param1  the first message parameter
     * @param param2  the second message parameter
     * @exception UnpositionedError this error will be positioned soon
     */
    public void fail(MessageDescription mess, Object param1, Object param2)
        throws UnpositionedError
    {
        throw new UnpositionedError(mess, param1, param2);
    }

    /**
     * Verify an expression and if false add an error
     * @param expr  condition to verify
     * @param phyl  where the error occur
     * @param mess  the message to be displayed
     * @return the value of expr
     * @exception UnpositionedError this error will be positioned soon
     */
    public boolean check(boolean expr,
                         MessageDescription mess)
        throws UnpositionedError
    {
        if (!expr) {
            fail(mess, null, null);
        }
        return expr;
    }

    /**
     * Verify an expression and if false add an error
     * @param expr  condition to verify
     * @param mess  the message to be displayed
     * @param param1  the first parameter
     * @return the value of expr
     * @exception UnpositionedError this error will be positioned soon
     */
    public boolean check(boolean expr,
                         MessageDescription mess,
                         Object param1)
        throws UnpositionedError
    {
        if (!expr) {
            fail(mess, param1, null);
        }
        return expr;
    }

    /**
     * Verify an expression and if false add an error
     * @param expr  condition to verify
     * @param mess  the message to be displayed
     * @param param1  the first parameter
     * @param param2  the second parameter
     * @return the value of expr
     * @exception UnpositionedError this error will be positioned soon
     */
    public boolean check(boolean expr,
                         MessageDescription mess,
                         Object param1,
                         Object param2)
        throws UnpositionedError
    {
        if (!expr) {
            fail(mess, param1, param2);
        }
        return expr;
    }

    /**
     * Gets the compiler
     */
    public Compiler getCompiler()
    {
        return parent.getCompiler();
    }

    protected CContext parent;
    protected CCompilationUnitContext cunit;
}
