/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

/**
 * This class represents a local variable information during check.
 */
public final class CVariableInfo
{
    /**
     * INF_INITIALIZED bitmask.
     */
    private static final int INF_INITIALIZED = 0x01;

    /**
     * INF_MAYBE_INITIALIZED bitmask.
     */
    private static final int INF_MAYBE_INITIALIZED = 0x02;

    /**
     * How many bits fit into a unit.
     */
    private static final int BITS_PER_UNIT = 2;

    /**
     * An array of bit masks.
     */
    private static final int[] MASKS = {
        3 << (2 * 0),
        3 << (2 * 1),
        3 << (2 * 2),
        3 << (2 * 3),
        3 << (2 * 4),
        3 << (2 * 5),
        3 << (2 * 6),
        3 << (2 * 7),
        3 << (2 * 8),
        3 << (2 * 9),
        3 << (2 * 10),
        3 << (2 * 11),
        3 << (2 * 12),
        3 << (2 * 13),
        3 << (2 * 14),
        3 << (2 * 15)
    };

    /**
     * The inverse bitmasks of the <code>MASKS</code> array.
     */
    private static final int[] REVMASKS = {
        ~(3 << (2 * 0)),
        ~(3 << (2 * 1)),
        ~(3 << (2 * 2)),
        ~(3 << (2 * 3)),
        ~(3 << (2 * 4)),
        ~(3 << (2 * 5)),
        ~(3 << (2 * 6)),
        ~(3 << (2 * 7)),
        ~(3 << (2 * 8)),
        ~(3 << (2 * 9)),
        ~(3 << (2 * 10)),
        ~(3 << (2 * 11)),
        ~(3 << (2 * 12)),
        ~(3 << (2 * 13)),
        ~(3 << (2 * 14)),
        ~(3 << (2 * 15))
    };

    /**
     * An array with variable infos.
     */
    private int[] infos;

    public CVariableInfo()
    {
        infos = new int[1];
    }
  
    private CVariableInfo(int[] infos)
    {
        this.infos = infos;
    }

    public Object clone()
    {
        return new CVariableInfo((int[])infos.clone());
    }

    public int getInfo(int pos)
    {
        int subscript = subscript(pos); // XXX: 'sub'
        int subset = subset(pos); // XXX: 'set'

        if (subscript < infos.length) {
            return ( (infos[subscript] & MASKS[subset])
                     >> (subset * BITS_PER_UNIT) );
        }
        else {
            return 0;
        }
    }

    public void setInfo(int pos, int info)
    {
        // System.err.println("setInfo(" + pos + ", " + info + ")");
        int subscript = subscript(pos);
        int subset = subset(pos);

        if (subscript >= infos.length) {
            int[] temp = new int[subscript + 1];

            System.arraycopy(infos, 0, temp, 0, infos.length);
            infos = temp;
        }

        infos[subscript] |= (info << (subset * BITS_PER_UNIT));
    }

    public void setInfo(int pos, int info, boolean forced)
    {
        int subscript = subscript(pos);
        int subset = subset(pos);

        if (subscript >= infos.length) {
            int[] temp = new int[subscript + 1];

            System.arraycopy(infos, 0, temp, 0, infos.length);
            infos = temp;
        }

        infos[subscript] = ( (info << (subset * BITS_PER_UNIT))
                             | (infos[subscript] & REVMASKS[subset]) );
    }

    /**
     * Merge two infos.
     *
     * @param info1
     * The first JLocalVariable info.
     * @param info2
     * The second JLocalVariable info.
     *
     * @return
     * The merged result of these information flags.
     */
    public static final int merge(int info1, int info2)
    {
        return ( ((info1 | info2) & INF_MAYBE_INITIALIZED)
                 | ((info1 & info2) & INF_INITIALIZED) );
    }

    /**
     * Initialize.
     */
    public static final int initialize()
    {
        return INF_INITIALIZED | INF_MAYBE_INITIALIZED;
    }

    /**
     * isInitialized.
     */
    public static final boolean isInitialized(int info)
    {
        return (info & INF_INITIALIZED) != 0;
    }

    /**
     * mayBeInitialized.
     */
    public static final boolean mayBeInitialized(int info)
    {
        return (info & INF_MAYBE_INITIALIZED) != 0;
    }

    /**
     * ...
     */
    private int subscript(int pos)
    {
        return pos >> 4;
    }

    /**
     * ...
     */
    private int subset(int pos)
    {
        return pos % 16;
    }
}
