/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package at.dms.kjc;

import at.dms.classfile.MultiarrayInstruction;
import at.dms.classfile.ClassRefInstruction;
import at.dms.classfile.NewarrayInstruction;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;

/**
 * This class implements a list of expression used in array initialisation
 * such as: <pre> int a[1][2][3]</pre>
 */
public class JArrayDimsAndInits
    extends JPhylum
{

    /**
     * Constructor
     * arr: brackets
     */
    public JArrayDimsAndInits(TokenReference where,
                              JExpression[] dims,
                              JArrayInitializer init)
    {
        super(where);

        this.init = init;
        this.dims = dims;
    }

    /**
     * Check expression and evaluate and alter context
     * @param context   the actual context of analyse
     * @return a pure java expression including promote node
     * @exception PositionedError Error catched as soon as possible
     * @exception UnpositionedError this error will be positioned soon
     */
    public CType checkExpression(CExpressionContext context, CType type)
        throws PositionedError
    {
        boolean expr = false;
        boolean hole = true;

        for (int i = 0; i < dims.length; i++) {
            if (dims[i] != null) {
                expr = true;
                dims[i] = dims[i].checkExpression(context);
                check(context,
                      dims[i].getType().implicitlyCastTo(CStdType.Integer),
                      KjcMessages.ARRAY_BADTYPE, dims[i].getType());
                hole = false;
            }
        }

        check(context, !hole || init != null, KjcMessages.ARRAY_NEW_NOSIZE);

        if (this.type == null) {
            this.type = new CArrayType(type, dims.length);
        }

        if (init != null) {
            check(context, !expr, KjcMessages.ARRAY_NEW_NOSIZE);
            init.setType(this.type);
            init.checkExpression(context);
        }
        return this.type;
    }

    /**
     * Accepts the specified visitor
     * @param p  the visitor
     */
    public void accept(KjcVisitor p)
    {
        p.visitArrayDimsAndInit(this, dims, init);
    }

    /**
     * Generates a sequence of bytecodes
     * @param code  the code list
     */
    public void genCode(CodeSequence code)
    {
        code.setLineNumber(getTokenReference().getLine());

        if (code.discardValue()) {
            return;
        }

        if (init == null) {
            allocArray(code, type, dims);
        }
        else {
            init.genCode(code);
        }
    }


    /**
     * Generates a sequence of bytecodes
     * @param code  the code list
     */
    public static void allocArray(CodeSequence code,
                                  CArrayType type, JExpression[] dims)
    {
        for (int i = 0; i < dims.length; i++) {
            if (dims[i] != null) {
                dims[i].genCode(code);
            }
        }

        if (type.getArrayBound() > 1) {
            int  count = 0;
            for (int i = 0; i < dims.length && dims[i] != null; i++) {
                count++;
            }

            code.plantInstruction
                (new MultiarrayInstruction(type.getSignature(), count));
        }
        else {
            if (type.getIndexedType().isReference()) {
                code.plantInstruction(new ClassRefInstruction
                    (opc_anewarray,
                     ((CClassType)type.getIndexedType()).getQualifiedName()));
            }
            else {
                byte num;

                switch (type.getIndexedType().getTypeID()) {
                case TID_BYTE:
                    num = 8;
                    break;
                case TID_BOOLEAN:
                    num = 4;
                    break;
                case TID_CHAR:
                    num = 5;
                    break;
                case TID_SHORT:
                    num = 9;
                    break;
                case TID_INT:
                    num = 10;
                    break;
                case TID_LONG:
                    num = 11;
                    break;
                case TID_FLOAT:
                    num = 6;
                    break;
                case TID_DOUBLE:
                    num = 7;
                    break;
                default:
                    fail();
                    num = -1;
                }
                code.plantInstruction(new NewarrayInstruction(num));
            }
        }
    }

    private CArrayType type;
    private JExpression[] dims;
    private JArrayInitializer init;
}
