/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package at.dms.kjc;

import at.dms.classfile.NoArgInstruction;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;
import at.dms.util.MessageDescription;

/**
 * Representation of bitwise and logical operations: binary and, binary or and
 * binary exclusive-or, as well as logical and, logical or and
 * logical exclusive-or. However, this does not include the shortcut logical
 * and and or operations. See JLS2 15.22.
 *
 * @see JConditionalAndExpression
 * @see JConditionalOrExpression
 *
 * @author DMS, Erwin Bolwidt
 */
public class JBitwiseExpression
    extends JBinaryArithmeticExpression
{

    /**
     * Construct a node in the parsing tree.
     *
     * @param where the line of this node in the source code
     * @param oper the operator
     * @param left the left operand
     * @param right the right operand
     */
    public JBitwiseExpression(TokenReference where,
                              int oper, JExpression left, JExpression right)
    {
        super(where, left, right);
        this.oper = oper;
    }

    /**
     * Check expression and evaluate and alter context.
     *
     * @param context the context for the check.
     * @return a pure java expression including promote node.
     *
     * @exception PositionedError
     * Any error in checking or compiling, with position information.
     */
    public JExpression checkBinaryBody(CExpressionContext context)
        throws PositionedError
    {
        try {
            type = computeType(left.getType(), right.getType(),
                               KjcMessages.BITWISE_BADTYPE);

            if (left.isConstant() && right.isConstant()) {
                return constantFolding();
            }
            else {
                return this;
            }
        }
        catch (UnpositionedError e) {
            throw e.addPosition(getTokenReference());
        }
    }

    /**
     * See superclass javadoc.
     */
    public static CType computeType(CType left, CType right,
                                    MessageDescription badTypeMessage)
        throws UnpositionedError
    {
        if (left == CStdType.Boolean && right == CStdType.Boolean) {
            return CStdType.Boolean;
        }
        else if (left.isOrdinal() && right.isOrdinal()) {
            return JBinaryArithmeticExpression.computeType
                (left, right, badTypeMessage);
        }
        else {
            throw new UnpositionedError(badTypeMessage, left, right);
        }
    }


    /**
     * Computes the result of the operation at compile-time (JLS 15.27).
     * @param left  the left value
     * @param right  the right value
     * @return a literal resulting of an operation over two literals
     */
    public JExpression constantFolding()
        throws UnpositionedError
    {
        CType leftType = left.getType();
        CType rightType = right.getType();

        if (leftType == CStdType.Boolean && rightType == CStdType.Boolean) {
            boolean result;

            result = compute(left.getBooleanLiteral().booleanValue(),
                             right.getBooleanLiteral().booleanValue());
            return new JBooleanLiteral(getTokenReference(), result);
        }
        else {
            return super.constantFolding();
        }
    }

    /**
     * Computes the result of the operation at compile-time (JLS 15.27).
     * @param left  the first operand
     * @param right  the seconds operand
     * @return the result of the operation
     */
    public int compute(int left, int right)
    {
        switch (oper) {
        case OPE_BAND:
            return left & right;
        case OPE_BOR:
            return left | right;
        case OPE_BXOR:
            return left ^ right;
        default:
            fail();
            return 0;
        }
    }

    /**
     * Computes the result of the operation at compile-time (JLS 15.27).
     * @param left  the first operand
     * @param right  the seconds operand
     * @return the result of the operation
     */
    public long compute(long left, long right)
    {
        switch (oper) {
        case OPE_BAND:
            return left & right;
        case OPE_BOR:
            return left | right;
        case OPE_BXOR:
            return left ^ right;
        default:
            fail();
            return 0L;
        }
    }

    /**
     * Computes the result of the operation at compile-time (JLS 15.27).
     * @param left  the first operand
     * @param right  the seconds operand
     * @return the result of the operation
     */
    public boolean compute(boolean left, boolean right)
    {
        switch (oper) {
        case OPE_BAND:
            return left & right;
        case OPE_BOR:
            return left | right;
        case OPE_BXOR:
            return left ^ right;
        default:
            fail();
            return false;
        }
    }

    /**
     * Cannot perform bitwise operation on floats.
     */
    public float compute(float left, float right)
    {
        fail();
        return 0f;
    }

    /**
     * Cannot perform bitwise operation on doubles.
     */
    public double compute(double left, double right)
    {
        fail();
        return 0d;
    }

    /**
     * Accepts the specified visitor
     * @param p the visitor
     */
    public void accept(KjcVisitor p)
    {
        p.visitBitwiseExpression(this, oper, left, right);
    }

    /**
     * Returns the JVM opcode for this operation.
     *
     * @param type the type of result.
     * @return the opcode for this operation.
     */
    public static int getOpcode(int oper, CType type)
    {
        if (type == CStdType.Long) {
            /* Select special opcodes for long operands */
            switch (oper) {
            case OPE_BAND:
                return opc_land;
            case OPE_BOR:
                return opc_lor;
            case OPE_BXOR:
                return opc_lxor;
            }
        }
        else {
            /* Select opcodes for integer _and_ boolean operands */
            switch (oper) {
            case OPE_BAND:
                return opc_iand;
            case OPE_BOR:
                return opc_ior;
            case OPE_BXOR:
                return opc_ixor;
            }
        }
        return -1;
    }

    /**
     * Generates a sequence of bytecodes.
     * @param code the code list
     */
    public void genCode(CodeSequence code)
    {
        code.setLineNumber(getTokenReference().getLine());

        code.setDiscardValue(false);

        left.genCode(code);
        right.genCode(code);

        code.plantInstruction
            (new NoArgInstruction(getOpcode(oper, getType())));
    }

    protected int oper;
}
