/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import at.dms.classfile.PushLiteralInstruction;
import at.dms.compiler.TokenReference;

/**
 * A boolean literal (true or false.)
 */
public class JBooleanLiteral
    extends JLiteral
{
    /**
     * Boolen value of this literal.
     */
    private boolean value;

    /**
     * Construct a node in the parsing tree.
     *
     * @param where
     * The location of this node in the source code.
     *
     * @param image
     * The string representation of this literal.
     */
    public JBooleanLiteral(TokenReference where, boolean value)
    {
        super(where);
        this.value = value;
    }

    /**
     * Compute the type of this expression (called after parsing.)
     *
     * @return the type of this expression.
     */
    public CType getType()
    {
        return CStdType.Boolean;
    }

    /**
     * Get a Number that represents the value of this literal.
     * Do not call before checkBody.
     */
    public boolean booleanValue()
    {
        return value;
    }

    /**
     * Whether this literal has the value that is the default for this type
     * of literal, meaning that this returns true of the value of this boolean
     * literal is false.
     */
    public boolean isDefault()
    {
        return booleanValue() == false;
    }

    /**
     * Get a Number that represents the value of this literal.
     * Do not call before checkBody.
     */
    public Object getValue()
    {
        return new Boolean(value);
        // return new Integer(value ? 1 : 0);
    }

    /**
     * Set the value of this literal.
     *
     * @param newValue
     * The new value.
     */
    public void setValue(boolean newValue)
    {
        value = newValue;
    }

    /**
     * Check expression and evaluate and alter context.
     *
     * @param context
     * The actual context of analyse.
     * @return
     * A pure java expression including promote node.
     *
     * @exception PositionedError
     * If the check fails.
     */
    public JExpression checkExpression(CExpressionContext context)
    {
        return this;
    }

    /**
     * Accepts the specified visitor.
     *
     * @param p
     * The visitor.
     */
    public void accept(KjcVisitor p)
    {
        p.visitBooleanLiteral(value);
    }

    /**
     * Generates a sequence of bytecodes.
     *
     * @param code
     * The code list.
     */
    public void genCode(CodeSequence code)
    {
        code.setLineNumber(getTokenReference().getLine());

        code.plantInstruction(new PushLiteralInstruction(value ? 1 : 0));
    }

    /**
     * Generates a sequence of bytecodes to branch on a label.
     * This method helps to handle heavy optimizables conditions.
     *
     * @param code
     * The code list.
     */
    public void genBranch(boolean cond, CodeSequence code, CodeLabel label)
    {
        code.setLineNumber(getTokenReference().getLine());

        if (value == cond) {
            code.plantLabelRef(opc_goto, label);
        }
    }
}
