/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package at.dms.kjc;

import at.dms.compiler.NumberParser;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;

/**
 * A simple character constant.
 */
public class JCharLiteral
    extends JOrdinalLiteral
{
    /**
     * Construct a node in the parsing tree.
     *
     * @param where
     * The line of this node in the source code.
     *
     * @param image
     * The string representation of this literal.
     */
    public JCharLiteral(TokenReference where, String image)
    {
        super(where, image);
        setType(CStdType.Char);
    }

    /**
     * Construct a node in the parsing tree.
     *
     * @param where
     * The line of this node in the source code.
     *
     * @param image
     * The string representation of this literal.
     */
    public JCharLiteral(TokenReference where, char val)
    {
        super(where, "");

        setType(CStdType.Char);
        setValue(new Integer(val));
        image = null;
        setType(CStdType.Char);
    }

    /**
     * Construct a node in the parsing tree
     * @param where  the line of this node in the source code
     * @param image  the string representation of this literal
     */
    public JCharLiteral(TokenReference where, Object val)
    {
        super(where, "");
        setType(CStdType.Char);
        if (val instanceof Character) {
            setValue(new Integer(((Character)val).charValue()));
        }
        else {
            setValue((Number)val);
        }
        image = null;
        setType(CStdType.Char);
    }

    /**
     * Construct a node in the parsing tree
     * @param where  the line of this node in the source code
     * @param image  the string representation of this literal
     * @param quoted  there is quote around image
     */
    public JCharLiteral(TokenReference where, String image, boolean quoted)
    {
        super(where, image.substring(1, image.length() - 1));
        setType(CStdType.Char);
    }

    /**
     * get a Number that represents the value of this literal
     * Do not call before checkBody
     */
    public Object getValue()
    {
        return new Character((char)getNumberValue().intValue());
    }

    /**
     * Check expression and evaluate and alter context
     * @param context the actual context of analyse
     * @return  a pure java expression including promote node
     * @exception PositionedError  if the check fails
     */
    public JExpression checkExpression(CExpressionContext context)
        throws PositionedError
    {
        if (image == null) {
            return this;
        }
        if (image.startsWith("\\u")) {
            setValue(new Integer((char)NumberParser.decodeHexInt
                                 (false, image.substring(2))));
        }
        else if (image.startsWith("\\") && image.length() > 1) {

            if ((image.charAt(1) >= '0') && (image.charAt(1) <= '7')) {
                setValue(new Integer((char)NumberParser.decodeOctInt
                                     (false, image.substring(1))));
                /*
                check(context,
                      charValue.charValue() <= 0377,
                      KjcMessages.INVALID_OCTAL_CHAR, image);
                */
            }
            else if (image.equals("\\b")) {
                setValue(new Integer('\b'));
            }
            else if (image.equals("\\r")) {
                setValue(new Integer('\r'));
            }
            else if (image.equals("\\t")) {
                setValue(new Integer('\t'));
            }
            else if (image.equals("\\n")) {
                setValue(new Integer('\n'));
            }
            else if (image.equals("\\f")) {
                setValue(new Integer('\f'));
            }
            else if (image.equals("\\\"")) {
                setValue(new Integer('\"'));
            }
            else if (image.equals("\\\'")) {
                setValue(new Integer('\''));
            }
            else if (image.equals("\\\\")) {
                setValue(new Integer('\\'));
            }
            else {
                check(context,
                      false,
                      KjcMessages.INVALID_ESCAPE_SEQUENCE, image);
            }
        }
        else {
            setValue(new Integer(image.charAt(0)));
        }

        image = null;
        return this;
    }

    /**
     * convertType
     * changes the type of this expression to an other
     * @param  dest the destination type
     */
    public JExpression convertType(CType dest, CExpressionContext context)
        throws PositionedError
    {
        if (dest == CStdType.Char) {
            return this;
        }
        else if (dest == CStdType.Byte) {
            return new JOrdinalLiteral(getTokenReference(),
                                       new Byte(getNumberValue().byteValue()),
                                       CStdType.Byte);
        }
        else if (dest == CStdType.Short) {
            return new JOrdinalLiteral(getTokenReference(),
                                       new Short
                                           (getNumberValue().shortValue()),
                                       CStdType.Short);
        }
        else if (dest == CStdType.Integer) {
            return new JOrdinalLiteral(getTokenReference(),
                                       new Integer
                                           (getNumberValue().intValue()),
                                       CStdType.Integer);
        }
        else if (dest == CStdType.Long) {
            return new JOrdinalLiteral(getTokenReference(),
                                       new Long(getNumberValue().longValue()),
                                       CStdType.Long);
        }
        else if (dest == CStdType.Float) {
            return new JRealLiteral(getTokenReference(),
                                    new Float(getNumberValue().floatValue()),
                                    CStdType.Float);
        }
        else if (dest == CStdType.Double) {
            return new JRealLiteral(getTokenReference(),
                                    new Double(getNumberValue().doubleValue()),
                                    CStdType.Double);
        }
        else {
            fail(dest.toString());
            return null;
        }
    }

    /**
     * Can this expression be converted to the specified type by assignment
     * conversion (JLS 5.2) ?
     * @param dest  the destination type
     * @return true iff the casting conversion is valid
    public boolean isAssignableTo(CType dest)
    {
    // Why?? The superclass does it well, and this one doesn't. EJB
        return CStdType.Char.isAssignableTo(dest);
    }
     */

    /**
     * Accepts the specified visitor
     * @param p  the visitor
     */
    public void accept(KjcVisitor p)
    {
        if (getNumberValue() != null) {
            p.visitCharLiteral((char)getNumberValue().intValue());
        }
        else {
            p.visitCharLiteral(image.charAt(0)); // !!! false
        }
    }
}
