/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package at.dms.kjc;

import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;

/**
 * JLS 14.2: Block
 *
 * TA block is a sequence of statements and local variable declaration
 * statements within braces.
 */
public class JConstructorBlock extends JBlock
{

    /**
     * Construct a node in the parsing tree
     * @param where  the line of this node in the source code
     */
    public JConstructorBlock(TokenReference where,
                             JStatement[] body)
    {
        super(where, body, null);
    }

    /**
     * Check statement.
     * @param context  the actual context of analyse
     * @param head  some declarations that take place in block
     *    such as int foo(int i) {}
     * @exception PositionedError  if the check fails
     */
    public void checkStatement(CBodyContext context) throws PositionedError
    {
        CBodyContext inside = context; // !!!context.createBodyContext();

        sourceClass = (CSourceClass)context.getClassContext().getCClass();
        if (explicitSuper == null
            && (body.length == 0
                || !(body[0] instanceof JExpressionStatement
                     && ((JExpressionStatement)body[0]).getExpression()
                     instanceof JExplicitConstructorInvocation))) {

            if (context.getClassContext().getCClass().getType()
                .equals(CStdType.Object)) {

                explicitSuper = null;
            }
            else {
                explicitSuper =
                    new JExplicitConstructorInvocation(getTokenReference(),
                                                       null,
                                                       "super",
                                                       JExpression.EMPTY);
            }
        }
        else if (body.length > 0
                 && body[0] instanceof JExpressionStatement
                 && ((JExpressionStatement)body[0]).getExpression()
                 instanceof JExplicitConstructorInvocation) {

            explicitSuper = (JExplicitConstructorInvocation)
                ((JExpressionStatement)body[0]).getExpression();
            body[0] = new JEmptyStatement(getTokenReference(), null);
        }

        if (sourceClass.isInnerClass()) {
            paramsLength =
                context.getMethodContext().getCMethod().getParameters().length;
        }

        if (context.getClassContext().hasInitializer()
            && !(explicitSuper.isThisInvoke())) {
            // "Block$();" ===>
            JNameExpression ident = new JNameExpression(getTokenReference(),
                                                        JAV_INIT);
            JExpression expr = new JMethodCallExpression(getTokenReference(),
                                                         ident,
                                                         JExpression.EMPTY);
            blockCall = new JExpressionStatement(getTokenReference(),
                                                 expr, null);
            // "Block$();" <===
        }

        if (explicitSuper != null) {
            explicitSuper.checkExpression(new CExpressionContext(inside));
            if (explicitSuper.isThisInvoke()) {
                context.getClassContext().markAllFieldToInitialized();
            }
        }

        if (blockCall != null) {
            blockCall.checkStatement(inside);
        }

        super.checkStatement(inside);

        inside.close(getTokenReference());
    }

    /**
     * Accepts the specified visitor
     * @param p  the visitor
     */
    public void accept(KjcVisitor p)
    {
        JStatement[] newBody = body;
        /* Explicit super invocations are removed from the statement array
         * internally, but for printing it needs to be inserted in the
         * array given to the pretty printing routine.
         * That's what this code does.
         */
        if (explicitSuper != null) {
            if (body[0] instanceof JEmptyStatement) {
                newBody = new JStatement[body.length];
                System.arraycopy(body, 0, newBody, 0, body.length);
            }
            else {
                newBody = new JStatement[body.length + 1];
                System.arraycopy(body, 0, newBody, 1, body.length);
            }
            newBody[0] =
                new JExpressionStatement(explicitSuper.getTokenReference(),
                                         explicitSuper, null);
        }
        p.visitBlockStatement(this, newBody, getComments());
    }

    /**
     * Generates a sequence of bytecodes
     * @param code  the code list
     */
    public void genCode(CodeSequence code) {
        code.setLineNumber(getTokenReference().getLine());

        if (explicitSuper != null) {
            explicitSuper.genCode(code);
        }
        else if (body.length == 0) {
            new JReturnStatement(getTokenReference(), null, null)
                .genCode(code);
        }

        if (sourceClass.isInnerClass()) {
            sourceClass.genInit(code, paramsLength);
        }

        if (blockCall != null) {
            blockCall.genCode(code);
        }

        for (int i = 0; i < body.length; i++) {
            body[i].genCode(code);
        }
    }

    private JExplicitConstructorInvocation explicitSuper;
    private JStatement blockCall;
    private CSourceClass sourceClass;
    private int paramsLength;
}
