/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package at.dms.kjc;

import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.JavaStyleComment;

/**
 * JLS 14.14: Continue Statement
 *
 * A continue statement may occur only in a while, do, or for statement;
 * statements of these three kinds are called iteration statements.
 * Control passes to the loop-continuation point of an iteration statement.
 */
public class JContinueStatement
    extends JStatement
{

    /**
     * Construct a node in the parsing tree
     * @param where  the line of this node in the source code
     * @param label  the label of the enclosing labeled statement
     */
    public JContinueStatement(TokenReference where,
                              String label,
                              JavaStyleComment[] comments)
    {
        super(where, comments);
        this.label = label;
    }

    /**
     * Check statement.
     *
     * @param context  the actual context of analyse
     * @exception PositionedError  if the check fails
     * @exception UnpositionedError this error will be positioned soon
     */
    public void checkStatement(CBodyContext context)
        throws PositionedError
    {
        if (label != null) {
            target = context.getLabeledStatement(label);
            check(context, target != null,
                  KjcMessages.LABEL_UNKNOWN, label);
            check(context, target.acceptsContinue(),
                  KjcMessages.CONTINUE_NOTLOOP);
        }
        else {
            target = context.getNearestContinuableStatement();
            check(context, target != null,
                  KjcMessages.CANNOT_CONTINUE);
        }

        context.addContinue();
        continueLabel = target.getContinueLabel();
    }

    /**
     * Accepts the specified visitor
     * @param p the visitor
     */
    public void accept(KjcVisitor p)
    {
        p.visitContinueStatement(this, label);
    }

    /**
     * Generates a sequence of bytecodes
     * @param code the code list
     */
    public void genCode(CodeSequence code)
    {
        code.setLineNumber(getTokenReference().getLine());

        code.plantBreak(target);

        code.plantLabelRef(opc_goto, continueLabel);

        continueLabel = null;
        target = null;
    }

    private CodeLabel continueLabel;
    private String label;
    private JStatement target;
}
