/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package at.dms.kjc;

import at.dms.compiler.CWarning;
import at.dms.compiler.JavaStyleComment;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;

/**
 * JLS 14.11: Do Statement
 *
 * The do statement executes an expression and a statement repeatedly
 * until the value of the expression is false.
 */
public class JDoStatement
    extends JLoopStatement
{

    /**
     * Construct a node in the parsing tree
     * @param where  the line of this node in the source code
     * @param condition  the expression to evaluate.
     * @param body  the loop body.
     */
    public JDoStatement(TokenReference where,
                        JExpression condition,
                        JStatement body,
                        JavaStyleComment[] comments)
    {
        super(where, comments);

        this.condition = condition;
        this.body = body;
    }

    /**
     * Check statement.
     *
     * @param context  the actual context of analyse
     * @exception PositionedError  if the check fails
     */
    public void checkStatement(CBodyContext context)
        throws PositionedError
    {
        try {
            CLoopContext self = context.createLoopContext(this);
            CBodyContext bodyContext = new CBodyContext(self);

            body.checkBody(bodyContext);
            bodyContext.close(getTokenReference());
            self.close(getTokenReference());

            condition =
                condition.checkExpression(context.createExpressionContext());
            check(context, condition.getType() == CStdType.Boolean,
                  KjcMessages.IF_COND_NOTBOOLEAN);

            if (condition instanceof JAssignmentExpression) {
                context.reportTrouble
                    (new CWarning(getTokenReference(),
                                  KjcMessages.ASSIGNMENT_IN_CONDITION));
            }
        }
        catch (CBlockError e) {
            context.reportTrouble(e);
        }

        if (condition.isConstant()) {
            if (condition.getBooleanLiteral().booleanValue()) {
                context.setReachable(false);
                if (hasBreak()) {
                    context.adopt(getBreakContextSummary());
                }
            }
        }
    }

    /**
     * Accepts the specified visitor
     * @param p the visitor
     */
    public void accept(KjcVisitor p)
    {
        super.accept(p);
        p.visitDoStatement(this, condition, body);
    }

    /**
     * Generates a sequence of bytecodes
     * @param code  the code list
     */
    public void genCode(CodeSequence code)
    {
        code.setLineNumber(getTokenReference().getLine());

        code.pushContext(this);

        CodeLabel  bodyLabel = new CodeLabel();

        code.plantLabel(bodyLabel);    // body:
        body.genCode(code);     //  BODY CODE
        code.plantLabel(getContinueLabel());   // cont:
        condition.genBranch(true, code, bodyLabel);  //  EXPR CODE; IFNE body
        code.plantLabel(getBreakLabel());    // end: ...

        code.popContext(this);
    }

    private JExpression condition;
    private JStatement body;
}
