/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package at.dms.kjc;

import at.dms.classfile.NoArgInstruction;
import at.dms.compiler.CWarning;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;

/**
 * JLS 15.20: Equality Operators ('==' and '!=')
 */
public class JEqualityExpression
    extends JBinaryExpression
{

    /**
     * Construct a node in the parsing tree
     * This method is directly called by the parser
     * @param where  the line of this node in the source code
     * @param oper  the operator; OPE_EQ for ==, other for !=
     * @param left  the left operand
     * @param right  the right operand
     */
    public JEqualityExpression(TokenReference where,
                               int oper,
                               JExpression left,
                               JExpression right)
    {
        super(where, left, right);
        this.oper = oper;
    }

    /**
     * Check expression and evaluate and alter context
     * @param context the actual context of analyse
     * @return  a pure java expression including promote node
     * @exception PositionedError  if the check fails
     * @exception UnpositionedError this error will be positioned soon
     */
    public JExpression checkExpression(CExpressionContext context)
        throws PositionedError
    {
        left = left.checkExpression(context);
        right = right.checkExpression(context);

        CType leftType = left.getType();
        CType rightType = right.getType();

        if (leftType.isNumeric()) {
            // JLS 15.20.1: Numerical Equality Operators
            check(context,
                  rightType.isNumeric(),
                  KjcMessages.EQUALITY_TYPE, leftType, rightType);

            promote(context);
        }
        else if (leftType == CStdType.Boolean) {
            // JLS 15.20.2: Boolean Equality Operators
            check(context,
                  rightType == CStdType.Boolean,
                  KjcMessages.EQUALITY_TYPE,
                  leftType, rightType);
            if (left instanceof JBooleanLiteral
                || right instanceof JBooleanLiteral) {

                context.reportTrouble
                    (new CWarning(getTokenReference(),
                                  KjcMessages.COMPARING_BOOLEAN_CONSTANT));
            }
        }
        else {
            // JLS 15.20.3: Reference Equality Operators
            check(context,
                  leftType.isReference() && rightType.isReference(),
                  KjcMessages.EQUALITY_TYPE, leftType, rightType);

            check(context,
                  rightType.isCastableTo(leftType)
                  || leftType.isCastableTo(rightType),
                  KjcMessages.EQUALITY_TYPE, leftType, rightType);

            if (left.getType().equals(CStdType.String)
                && right.getType().equals(CStdType.String)
                && (left.isConstant() || right.isConstant())) {

                context.reportTrouble
                    (new CWarning(getTokenReference(),
                                  KjcMessages.STRING_COMPARISON));      
            }
        }

        type = CStdType.Boolean;

        if (left.isConstant() && right.isConstant()) {
            return constantFolding(left.getLiteral(), right.getLiteral());
        }
        else {
            return this;
        }
    }

    /**
     * @return a literal resulting of an operation over two literals
     */
    public JExpression constantFolding(JLiteral left, JLiteral right)
    {
        CType leftType = left.getType();
        CType rightType = right.getType();

        if (leftType.isNumeric()) {
            Number leftNumber = left.getNumberLiteral().getNumberValue();
            Number rightNumber = right.getNumberLiteral().getNumberValue();
            boolean equals; 

            if (leftType == CStdType.Double) {
                equals = leftNumber.doubleValue() == rightNumber.doubleValue();
            }
            else if (leftType == CStdType.Float) {
                equals = leftNumber.floatValue() == rightNumber.floatValue();
            }
            else {
                equals = leftNumber.longValue() == rightNumber.longValue();
            }

            if (oper == OPE_EQ) {
                return new JBooleanLiteral(getTokenReference(), equals);
            }
            else {
                return new JBooleanLiteral(getTokenReference(), !equals);
            }
        }
        else if (leftType == CStdType.Boolean) {
            boolean leftBoolean = left.getBooleanLiteral().booleanValue();
            boolean rightBoolean = right.getBooleanLiteral().booleanValue();
            if (oper == OPE_EQ) {
                return new JBooleanLiteral
                    (getTokenReference(), leftBoolean == rightBoolean);
            }
            else {
                return new JBooleanLiteral
                    (getTokenReference(), leftBoolean != rightBoolean);
            }
        }
        else if (leftType.equals(CStdType.String)) {
            boolean equals = left.getStringLiteral().stringValue()
                .equals(right.getStringLiteral().stringValue());

            if (oper == OPE_EQ) {
                return new JBooleanLiteral(getTokenReference(), equals);
            }
            else {
                return new JBooleanLiteral(getTokenReference(), !equals);
            }
        }

        /* Note: no expression containing the 'null' literal is constant,
         * see JLS2 15.28. So don't put any special cases for 'null' in
         * this method!
         */

        return this;
    }

    /**
     * Accepts the specified visitor.
     * @param p  the visitor
     */
    public void accept(KjcVisitor p)
    {
        p.visitEqualityExpression(this, oper, left, right);
    }

    /**
     * Generates a sequence of bytecodes
     * @param code  the code list
     */
    public void genCode(CodeSequence code)
    {
        genBooleanResultCode(code);
    }

    /**
     * Optimize a bi-conditionalitional expression
     */
    protected void genBranch(JExpression left,
                             JExpression right,
                             boolean conditional,
                             CodeSequence code,
                             CodeLabel label)
    {
        code.setLineNumber(getTokenReference().getLine());

        if ((left.isConstant() || right.isConstant())) {
            if (left instanceof JNullLiteral
                || right instanceof JNullLiteral) {

                if (left instanceof JNullLiteral) {
                    JExpression temp = left;
                    left = right;
                    right = temp;
                }
                left.genCode(code);
                code.plantLabelRef("equalityExpr: genBranch/1",
                                   conditional == (oper == OPE_EQ)
                                   ? opc_ifnull : opc_ifnonnull, label);
                return;
            }

            // Check if this expression checks for equality with 0.
            else if (((left instanceof JOrdinalLiteral
                       &&
                       ((JOrdinalLiteral)left).getNumberValue().intValue() == 0)
                      ||
                      (right instanceof JOrdinalLiteral
                       &&
                       ((JOrdinalLiteral)right).getNumberValue().intValue() == 0)
                      )
                     && left.getType().isCodedHasInt()) {

                if ((left instanceof JOrdinalLiteral)
                    && ((JOrdinalLiteral)left).getNumberValue().intValue() == 0) {

                    JExpression temp = left;
                    left = right;
                    right = temp;
                }

                left.genCode(code);
                code.plantLabelRef("equalityExpr: genBranch/2/"
                                   + (oper == OPE_EQ ? "==" : "!="),
                                   conditional == (oper == OPE_EQ)
                                   ? opc_ifeq : opc_ifne, label);
                return;
            }
        }

        left.genCode(code);
        right.genCode(code);

        switch (left.getType().getTypeID()) {
        case TID_ARRAY:
        case TID_CLASS:
            code.plantLabelRef("equalityExpr: genBranch/3",
                               conditional  == (oper == OPE_EQ)
                               ? opc_if_acmpeq : opc_if_acmpne, label);
            return;
        case TID_FLOAT:
            code.plantInstruction(new NoArgInstruction(opc_fcmpl));
            break;
        case TID_LONG:
            code.plantInstruction(new NoArgInstruction(opc_lcmp));
            break;
        case TID_DOUBLE:
            code.plantInstruction(new NoArgInstruction(opc_dcmpl));
            break;
        default:
            code.plantLabelRef("equalityExpr: genBranch/4",
                               conditional == (oper == OPE_EQ)
                               ? opc_if_icmpeq : opc_if_icmpne, label);
            return;
        }

        code.plantLabelRef("equalityExpr: genBranch/5",
                           conditional == (oper == OPE_EQ)
                           ? opc_ifeq : opc_ifne, label);
    }

    protected final int oper;
}

