/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package at.dms.kjc;

import at.dms.classfile.MethodRefInstruction;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;

/**
 * This class represents a explicit call to a super or self constructor.
 */
public class JExplicitConstructorInvocation
    extends JExpression
{

    private CClass  clazz;
    private CMethod  method;
    private JExpression prefix;
    private String  ident;
    private JExpression[] params;

    /**
     * Construct a node in the parsing tree
     * This method is directly called by the parser
     * @param where the line of this node in the source code
     * @param prefix an expression that is a field of a class representing a
     *               method.
     * @param ident the method name
     * @param args the argument of the call
     */
    public JExplicitConstructorInvocation(TokenReference where,
                                          JExpression prefix,
                                          String ident,
                                          JExpression[] args)
    {
        super(where);

        this.prefix = prefix;
        this.ident = ident;
        this.params = args;
    }

    /**
     * Returns true if it's this() else it's super().
     */
    final boolean isThisInvoke()
    {
        return ident.equals(JAV_THIS);
    }

    /**
     *
     */
    public CType getType()
    {
        return null;
    }

    /**
     * Check expression and evaluate and alter context
     * @param context the actual context of analyse
     * @return  a pure java expression including promote node
     * @exception PositionedError Error catched as soon as possible
     * @exception UnpositionedError Error catched as soon as possible
     */
    public JExpression checkExpression(CExpressionContext context)
        throws PositionedError
    {
        ((CConstructorContext)context.getMethodContext())
            .setSuperConstructorCalled(false);

        // !!! check in constructor !!!

        CType[] argsType = new CType[params.length];
        for (int i = 0; i < argsType.length; i++) {
            params[i] = params[i].checkExpression(context);

            argsType[i] = params[i].getType();
            assert(argsType[i] != null);
        }

        CClass local = context.getClassContext().getCClass();
        if (isThisInvoke()) {
            clazz = context.getClassContext().getCClass();
        }
        else {
            clazz = context.getClassContext().getCClass().getSuperClass();
        }

        assert(clazz != null);
        try {
            method = clazz.lookupMethod(local, JAV_CONSTRUCTOR, argsType);
        }
        catch (UnpositionedError e) {
            throw e.addPosition(getTokenReference());
        }


        check(context,
              method != null,
              KjcMessages.CONSTRUCTOR_NOTFOUND, clazz.getIdent());

        if (method.getOwner() != clazz) {
            // May be an inner class
            if (clazz.isInnerClass()) {
                CType[] argsType2 = new CType[argsType.length + 1];
                System.arraycopy(argsType, 0, argsType2, 1, argsType.length);
                /* Outer-this is at index 0! */
                argsType2[0] = clazz.getOwner().getType();

                try {
                    method = clazz.lookupMethod(local, JAV_CONSTRUCTOR,
                                                argsType2);
                }
                catch (UnpositionedError e) {
                    throw e.addPosition(getTokenReference());
                }
            }
            if (method.getOwner() != clazz) {
                // do not want a super constructor !
                throw new CMethodNotFoundError(getTokenReference(),
                                               null,
                                               clazz.getType().toString(),
                                               argsType);
            }
        }

        check(context,
              method.isAccessible(local),
              KjcMessages.METHOD_NOACCESS, method.getIdent());

        check(context,
              method.getOwner().isAccessible(local),
              KjcMessages.CLASS_NOACCESS, method.getOwner().getIdent());

        CClassType[] exceptions = method.getThrowables();
        for (int i = 0; i < exceptions.length; i++) {
            if (exceptions[i].isCheckedException()) {
                CBodyContext bodyContext = context.getBodyContext();
                CBodyContext clonedBodyContext = bodyContext.cloneContext();

                context.getBodyContext()
                    .addThrowable(new CThrowableInfo(exceptions[i], this,
                                                     clonedBodyContext));
            }
        }

        check(context,
              !context.getMethodContext().getCMethod().isStatic(),
              KjcMessages.BAD_THIS_STATIC);

        argsType = method.getParameters();
        for (int i = 0; i < params.length; i++) {
            params[i] = params[i].convertType(argsType[i], context);
        }

        ((CConstructorContext)context.getMethodContext())
            .setSuperConstructorCalled(true);

        return this;
    }

    /**
     * Accepts the specified visitor
     * @param p  the visitor
     */
    public void accept(KjcVisitor p)
    {
        p.visitExplicitConstructorInvocation(this, prefix, ident, params);
    }

    /**
     * Generates a sequence of bytecodes
     * @param code  the code list
     */
    public void genCode(CodeSequence code)
    {
        code.setLineNumber(getTokenReference().getLine());
        code.plantLoadThis();

        for (int i = 0; i < params.length; i++) {
            params[i].genCode(code);
        }

        if (method.getOwner().isInnerClass()) {
            clazz.genSyntheticParamsFromExplicitSuper(params, code);
        }

        code.plantInstruction(new MethodRefInstruction(opc_invokespecial,
                                                       method.getPrefixName(),
                                                       method.getIdent(),
                                                       method.getSignature()));
    }
}
