/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import at.dms.compiler.PositionedError;
import at.dms.compiler.CWarning;
import at.dms.compiler.TokenReference;
import at.dms.compiler.JavaStyleComment;

/**
 * JLS 14.10: If Statement
 *
 * The if statement executes an expression and a statement repeatedly
 * until the value of the expression is false.
 */
public class JIfStatement
    extends JStatement
{
    /**
     * ...
     */
    private JExpression condition;

    /**
     * ...
     */
    private JStatement thenClause;

    /**
     * ...
     */
    private JStatement elseClause;

    /**
     * Construct a node in the parsing tree.
     *
     * @param where  the line of this node in the source code.
     * @param cond  the expression to evaluate.
     * @param thenClause the statement to execute if the condition is true.
     * @param elseClause the statement to execute if the condition is false.
     */
    public JIfStatement(TokenReference where,
                        JExpression condition,
                        JStatement thenClause,
                        JStatement elseClause,
                        JavaStyleComment[] comments)
    {
        super(where, comments);

        this.condition = condition;
        this.thenClause = thenClause;
        this.elseClause = elseClause;
    }

    /**
     * Check statement.
     *
     * @param context  the actual context of analyse
     * @exception PositionedError  if the check fails
     * @exception UnpositionedError this error will be positioned soon
     */
    public void checkStatement(CBodyContext context)
        throws PositionedError
    {
        
        condition =
            condition.checkExpression(context.createExpressionContext());

        check(context,
              condition.getType() == CStdType.Boolean,
              KjcMessages.IF_COND_NOTBOOLEAN,
              condition.getType());

        CBodyContext thenContext = context.cloneContext();
        CBodyContext elseContext = context.cloneContext();

        thenClause.checkBody(thenContext);

        if (elseClause != null) {
            elseClause.checkBody(elseContext);
        }

        if ( condition.isConstant() ) {
            if (condition.getBooleanLiteral().booleanValue()) {
                // System.err.println("Constant, true");
                /* Condition is constant and true, so only assignment
                 * info from then context is valid.
                 */
                context.adopt(thenContext);
                if (elseClause != null) {
                    context.merge(elseContext);
                }
            }
            else {
                // System.err.println("Constant, false");
                /* Condition is constant and false, so only assignment
                 * info from else context is valid.
                 */

                // Difficult rule. adopt here seems to work. Wish that DMS
                // made more documentation..
                context.adopt(elseContext);
                context.merge(thenContext);
            }
        }
        else {
            if ( elseContext.isReachable() ) {
                if ( thenContext.isReachable() ) {
                    // System.err.println("NonConstant, elseReachable, thenReachable");
                    context.adopt(thenContext);
                    context.merge(elseContext);
                }
                else {
                    // System.err.println("NonConstant, elseReachable, thenUnreachable");
                    context.adopt(elseContext);
                }
            }
            else if (thenContext.isReachable() ) {
                // System.err.println("NonConstant, elseUnReachable, thenReachable");
                context.adopt(thenContext);
            }
            else {
                // System.err.println("NonConstant, elseUnReachable, thenUnReachable");
                context.setReachable(false);
            }
        }

        if (condition instanceof JAssignmentExpression) {
            context.reportTrouble
                (new CWarning(getTokenReference(),
                              KjcMessages.ASSIGNMENT_IN_CONDITION));
        }
    }

    /**
     * Accepts the specified visitor.
     *
     * @param p
     * The visitor.
     */
    public void accept(KjcVisitor p)
    {
        super.accept(p);
        p.visitIfStatement(this, condition, thenClause, elseClause);
    }

    /**
     * Generates a sequence of bytecodes.
     *
     * @param code
     * The code list.
     */
    public void genCode(CodeSequence code)
    {
        code.setLineNumber(getTokenReference().getLine());

        if (condition.isConstant()) {
            if (condition.getBooleanLiteral().booleanValue()) {
                thenClause.genCode(code);
            }
            else if (elseClause != null) {
                elseClause.genCode(code);
            }
        }
        else {
            CodeLabel  elseLabel = new CodeLabel();
            CodeLabel  nextLabel = new CodeLabel();

            //  COND IFEQ else
            condition.genBranch(false, code, elseLabel);
            //  THEN CODE
            thenClause.genCode(code);
            //  GOTO next
            code.plantLabelRef("if:GOTO next", opc_goto, nextLabel);
            // else:
            code.plantLabel(elseLabel);
            //  ELSE CODE
            if (elseClause != null) {
                elseClause.genCode(code);
            }
            // next ...
            code.plantLabel(nextLabel);
        }
    }
}
