/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;

/**
 * Root class for all literal expressions.
 *
 * By default, this all literals are constant
 * but there are circumstances in which this may be turned to false,
 * for example when KJC knows something is constant and can optimize
 * some things safely, but the JLS says that the result, even though a
 * literal, must not be considered a constant.
 */
public abstract class JLiteral
    extends JExpression
{
    /**
     * Construct a node in the parsing tree.
     *
     * @param where
     * The location of this node in the source code.
     */
    public JLiteral(TokenReference where)
    {
        super(where);
        setConstant(true);
    }

    /**
     * Create a literal from a type and a value.
     */
    public static JLiteral createLiteral(CType type, Object value)
    {
        if (type == CStdType.Char) {
            return new JCharLiteral(TokenReference.NO_REF, value);
        }
        else if (value instanceof Character) {
            value = new Integer(((Character)value).charValue());
        }

        if (type.isOrdinal()) {
            switch (type.getTypeID()) {
            case TID_BYTE:
                return new JOrdinalLiteral(TokenReference.NO_REF,
                                           ((Number)value).byteValue(),
                                           CStdType.Byte);
            case TID_SHORT:
                return new JOrdinalLiteral(TokenReference.NO_REF,
                                           ((Number)value).shortValue(),
                                           CStdType.Short);
            case TID_INT:
                return new JOrdinalLiteral(TokenReference.NO_REF,
                                           ((Number)value).intValue(),
                                           CStdType.Integer);
            case TID_LONG:
                return new JOrdinalLiteral(TokenReference.NO_REF,
                                           ((Number)value).longValue(),
                                           CStdType.Long);
            }
        }

        if (type == CStdType.Double) {
            return new JRealLiteral(TokenReference.NO_REF,
                                    (Number)value,
                                    CStdType.Double);
        }

        if (type == CStdType.Float) {
            return new JRealLiteral(TokenReference.NO_REF,
                                    (Number)value,
                                    CStdType.Float);
        }

        if (type == CStdType.Boolean) {
            boolean booleanValue;

            if (value instanceof Integer) {
                booleanValue = (((Integer)value).intValue() != 0);
            }
            else if (value instanceof Boolean) {
                booleanValue = ((Boolean)value).booleanValue();
            }
            else {
                fail();
                return null;
            }

            return new JBooleanLiteral(TokenReference.NO_REF, booleanValue);
        }

        if (type.equals(CStdType.String)) {
            return new JStringLiteral(TokenReference.NO_REF, value.toString());
        }

        fail();
        return null;
    }

    /**
     * Whether this expression is an lvalue: no, literals aren't.
     *
     * @return false
     */
    public boolean isLValue(CExpressionContext context)
    {
        return false;
    }

    /**
     * Returns true since all JLiteral instances and subclasses are literal.
     *
     * @return true if this expression is literal.
     */
    public boolean isLiteral()
    {
        return false;
    }

    /**
     * Returns the literal value of an expression.
     */
    public JLiteral getLiteral()
    {
        return this;
    }

    /**
     * Whether the value of this literal is the default value for literals
     * of this type.
     */
    public abstract boolean isDefault();

    /**
     * Returns the value of this literal.
     */
    public abstract Object getValue();

}
