/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package at.dms.kjc;

import at.dms.classfile.NoArgInstruction;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;

/**
 * A representation of the multiplication operator (JLS2 15.17.1)
 *
 * @author DMS, Erwin Bolwidt
 */
public strictfp class JMultExpression
    extends JBinaryArithmeticExpression
{

    /**
     * Construct a node in the parsing tree.
     *
     * @param where the line of this node in the source code
     * @param left the left operand
     * @param right the right operand
     */
    public JMultExpression(TokenReference where,
                           JExpression left, JExpression right)
    {
        super(where, left, right);
    }

    /**
     * Check expression and evaluate and alter context.
     *
     * @param context the actual context of analyse
     *
     * @return a pure java expression including promote node
     *
     * @exception PositionedError
     * Any compilation-related errror with position information.
     */
    public JExpression checkBinaryBody(CExpressionContext context)
        throws PositionedError
    {
        try {
            type = computeType(left.getType(), right.getType(),
                               KjcMessages.MULT_BADTYPE);

            if (left.isConstant() && right.isConstant()) {
                return constantFolding();
            }
            else {
                return this;
            }
        }
        catch (UnpositionedError e) {
            throw e.addPosition(getTokenReference());
        }

    }

    /**
     * Computes the result of the * operation on two integer constant values.
     *
     * @param left the left operand
     * @param right the right operand
     * @return the result of the operation
     */
    public int compute(int left, int right)
    {
        return left * right;
    }

    /**
     * Computes the result of the * operation on two long constant values.
     *
     * @param left the left operand
     * @param right the right operand
     * @return the result of the operation
     */
    public long compute(long left, long right)
    {
        return left * right;
    }

    /**
     * Computes the result of the * operation on two float constant values.
     *
     * @param left the left operand
     * @param right the right operand
     * @return the result of the operation
     */
    public float compute(float left, float right)
    {
        return left * right;
    }

    /**
     * Computes the result of the * operation on two double constant values.
     *
     * @param left the left operand
     * @param right the right operand
     * @return the result of the operation
     */
    public double compute(double left, double right)
    {
        return left * right;
    }

    /**
     * Accepts the specified visitor
     * @param p the visitor
     */
    public void accept(KjcVisitor p)
    {
        p.visitBinaryExpression(this, "*", left, right);
    }

    /**
     * Returns the opcode for multiplication, depending on the type of the
     * operands.
     *
     * @param type the type to which the operands are promoted.
     *
     * @return the opcode for this operation
     */
    public static int getOpcode(CType type)
    {
        switch (type.getTypeID()) {
        case TID_FLOAT:
            return opc_fmul;
        case TID_LONG:
            return opc_lmul;
        case TID_DOUBLE:
            return opc_dmul;
        default:
            return opc_imul;
        }
    }

    /**
     * Generates a sequence of bytecodes for this expression.
     *
     * @param code the code list
     */
    public void genCode(CodeSequence code)
    {
        code.setLineNumber(getTokenReference().getLine());

        left.genCode(code);
        right.genCode(code);

        code.plantInstruction(new NoArgInstruction(getOpcode(getType())));
    }
}
