/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package at.dms.kjc;

import at.dms.classfile.IincInstruction;
import at.dms.classfile.PushLiteralInstruction;
import at.dms.classfile.NoArgInstruction;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;

/**
 * This class represents prefix increment and decrement expressions.
 * 15.13.2 Prefix Increment Operator ++
 * 15.13.3 Prefix Decrement Operator --
 */
public class JPrefixExpression
    extends JExpression
{

    /**
     * Construct a node in the parsing tree
     * @param where  the line of this node in the source code
     * @param oper  the operator
     * @param expr  the operand
     */
    public JPrefixExpression(TokenReference where,
                             int oper,
                             JExpression expr)
    {
        super(where);
        this.oper = oper;
        this.expr = expr;
    }

    /**
     * Compute the type of this expression (called after parsing)
     * @return the type of this expression
     */
    public CType getType()
    {
        return type;
    }

    /**
     * Returns true iff this expression can be used as a statement (JLS 14.7)
     */
    public boolean isStatementExpression()
    {
        return true;
    }

    /**
     * Check expression and evaluate and alter context
     * @param context the actual context of analyse
     * @return  a pure java expression including promote node
     * @exception PositionedError  if the check fails
     * @exception UnpositionedError this error will be positioned soon
     */
    public JExpression checkExpression(CExpressionContext context)
        throws PositionedError
    {
        expr = expr.checkExpression(context, true, context.discardValue());
        check(context, expr.getType().isNumeric(),
              KjcMessages.PREFIX_BADTYPE, expr.getType());
        check(context,
              (expr instanceof CInitializable)
              && ((CInitializable)expr).isLValue(context),
              KjcMessages.PREFIX_NOTLVALUE);
        check(context,
              ((CInitializable)expr).isInitialized(context),
              KjcMessages.PREFIX_NOTINITIALIZED);

        type = expr.getType();

        return this;
    }

    /**
     * Accepts the specified visitor.
     * @param p  the visitor
     */
    public void accept(KjcVisitor p)
    {
        p.visitPrefixExpression(this, oper, expr);
    }

    /**
     * Generates a sequence of bytecodes.
     * @param code  the code list
     */
    public void genCode(CodeSequence code)
    {
        code.setLineNumber(getTokenReference().getLine());
        boolean  discardValue = code.discardValue();

        int   val = oper == OPE_PREINC ? 1 : -1;

        if ((expr.getType() == CStdType.Integer) &&
            (expr instanceof JLocalVariableExpression)) {
            code.plantInstruction
                (new IincInstruction(((JLocalVariableExpression)expr)
                                     .getPosition(),
                                     val));
            code.setDiscardValue(discardValue);
            expr.genCode(code);
            code.setDiscardValue(false);
        }
        else {
            code.setDiscardValue(false);

            ((CInitializable)expr).genStartStoreCode(code);

            expr.genCode(code);
            switch (expr.getType().getTypeID()) {
            case TID_FLOAT:
                code.plantInstruction(new PushLiteralInstruction((float)val));
                code.plantInstruction(new NoArgInstruction(opc_fadd));
                break;
            case TID_LONG:
                code.plantInstruction(new PushLiteralInstruction((long)val));
                code.plantInstruction(new NoArgInstruction(opc_ladd));
                break;
            case TID_DOUBLE:
                code.plantInstruction(new PushLiteralInstruction((double)val));
                code.plantInstruction(new NoArgInstruction(opc_dadd));
                break;

            case TID_BYTE:
            case TID_CHAR:
            case TID_SHORT:
            case TID_INT:
                code.plantInstruction(new PushLiteralInstruction(val));
                code.plantInstruction(new NoArgInstruction(opc_iadd));
                switch (expr.getType().getTypeID()) {
                case TID_BYTE:
                    code.plantInstruction(new NoArgInstruction(opc_i2b));
                    break;
                case TID_CHAR:
                    code.plantInstruction(new NoArgInstruction(opc_i2c));
                    break;
                case TID_SHORT:
                    code.plantInstruction(new NoArgInstruction(opc_i2s));
                }
            }

            ((CInitializable)expr).genEndStoreCode(code, discardValue);
        }
    }

    protected int oper;
    protected JExpression expr;
    protected CType type;
}
