/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.JavaStyleComment;
import at.dms.util.MessageDescription;

/**
 * The root class for all statements. (JLS 14.4)
 */
public abstract class JStatement
    extends JPhylum
{
    /**
     * Comments for statements.
     */
    private JavaStyleComment[] comments;

    /**
     * Construct a node in the parsing tree
     * @param where the line of this node in the source code
     */
    public JStatement(TokenReference where,
                      JavaStyleComment[] comments)
    {
        super(where);
        this.comments = comments;
    }

    /**
     * Check statement.
     *
     * @param context  the actual context of analyse
     * @exception PositionedError  if the check fails
     */
    public void checkBody(CBodyContext context)
        throws PositionedError
    {
        checkStatement(context);
    }

    /**
     * Check statement.
     *
     * @param context  the actual context of analyse
     * @exception PositionedError  if the check fails
     * @exception UnpositionedError this error will be positioned soon
     */
    public abstract void checkStatement(CBodyContext context)
        throws PositionedError;

    /**
     * Adds a compiler error.
     *
     * @param context  the context in which the error occurred
     * @param key  the message ident to be displayed
     * @param params  the array of parameters
     */
    protected void fail(CContext context,
                        MessageDescription key,
                        Object[] params)
        throws PositionedError
    {
        throw new CLineError(getTokenReference(), key, params);
    }

    /**
     * Returns true iff this statement is a possible target for a break
     * statement.
     * This method is overridden in classes that accept breaks.
     */
    public boolean acceptsBreak()
    {
        return false;
    }

    /**
     * Returns true iff this statement is a possible target for a continue
     * statement.
     * This method is overridden in classes that accept continues.
     */
    public boolean acceptsContinue()
    {
        return false;
    }

    /**
     * Returns a label at end of this statement (for break statement)
     */
    public CodeLabel getBreakLabel()
    {
        throw new RuntimeException("NO END LABEL");
    }

    /**
     * Returns the beginning of this block (for continue statement)
     */
    public CodeLabel getContinueLabel()
    {
        throw new RuntimeException("NO CONTINUE LABEL");
    }

    /**
     * Informs statement that there is at least one break.
     */
    public void addBreak(CBodyContext context)
    {
        throw new RuntimeException("CANNOT BREAK");
    }

    /**
     * Generates a sequence of bytecodes.
     *
     * @param code
     * The code list.
     */
    public abstract void genCode(CodeSequence code);

    /**
     * Accepts the specified visitor.
     *
     * @param p
     * The visitor.
     */
    public void accept(KjcVisitor p)
    {
        if (comments != null) {
            p.visitComments(comments);
        }
    }

    /**
     * Returns the comments.
     */
    public JavaStyleComment[] getComments()
    {
        return comments;
    }
}
