/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package at.dms.kjc;

import at.dms.classfile.NoArgInstruction;
import at.dms.classfile.HandlerInfo;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;
import at.dms.compiler.JavaStyleComment;

/**
 * JLS 14.16: Synchronized Statement
 *
 * A synchronized statement acquires a mutual-exclusion lock on behalf
 * of the executing thread, executes a block, then releases the lock.
 */
public class JSynchronizedStatement
    extends JStatement
{

    /**
     * Construct a node in the parsing tree
     * @param where  the line of this node in the source code
     * @param cond  the expression to evaluate.
     * @param body  the loop body.
     */
    public JSynchronizedStatement(TokenReference where,
                                  JExpression cond,
                                  JStatement body,
                                  JavaStyleComment[] comments)
    {
        super(where, comments);
        this.cond = cond;
        this.body = body;
    }

    /**
     * Check statement.
     *
     * @param context  the actual context of analyse
     * @exception PositionedError  if the check fails
     */
    public void checkStatement(CBodyContext context)
        throws PositionedError
    {
        localVar =
            new JGeneratedLocalVariable(null,
                                        0,
                                        CStdType.Integer,
                                        "sync" + toString() /* unique ID */,
                                        null);

        /* Fix-EJB: search for enclosing try/finally and define this var
         * inside its block; else this variable may be 'forgotten'!
         * That results in verifier errors.
         */
        CContext findTry = context;
        while ( findTry != null
                && !(findTry instanceof CTryContext)
                && !(findTry instanceof CClassContext) ) {

            findTry = findTry.getParentContext();
        }
 
        try {
            if ( findTry != null && findTry instanceof CTryContext ) {
                findTry.getBlockContext().addVariable(localVar);
            }
            else {
                context.getBlockContext().addVariable(localVar);
            }
        }
        catch (UnpositionedError e) {
            throw e.addPosition(getTokenReference());
        }
        cond = cond.checkExpression(context.createExpressionContext());
        check(context, cond.getType().isReference(),
              KjcMessages.SYNCHRONIZED_NOTREFERENCE);
        body.checkBody(context);
    }

    /**
     * Accepts the specified visitor
     * @param p  the visitor
     */
    public void accept(KjcVisitor p)
    {
        super.accept(p);
        p.visitSynchronizedStatement(this, cond, body);
    }

    /**
     * Generates a sequence of bytecodes
     * @param code  the code list
     */
    public void genMonitorExit(CodeSequence code)
    {
        code.plantLocalVar(opc_aload, localVar);
        code.plantInstruction(new NoArgInstruction(opc_monitorexit));
    }

    /**
     * Generates a sequence of bytecodes
     * @param code  the code list
     */
    public void genCode(CodeSequence code)
    {
        code.setLineNumber(getTokenReference().getLine());

        CodeLabel nextLabel = new CodeLabel();

        cond.genCode(code);
        code.plantInstruction(new NoArgInstruction(opc_dup));
        code.plantLocalVar(opc_astore, localVar);
        code.plantInstruction(new NoArgInstruction(opc_monitorenter));

        code.pushContext(this);
        int startPC = code.getPC();
        body.genCode(code);
        code.popContext(this);

        genMonitorExit(code);
        int endPC = code.getPC();
        code.plantLabelRef(opc_goto, nextLabel);

        int errorPC = code.getPC();
        genMonitorExit(code);
        code.plantInstruction(new NoArgInstruction(opc_athrow));

        code.plantLabel(nextLabel);   // next: ...

        // protect
        code.addExceptionHandler
            (new HandlerInfo(code.getInstructionAt(startPC),
                             code.getInstructionAt(endPC),
                             code.getInstructionAt(errorPC),
                             null));
    }

    private JExpression cond;
    private JStatement body;
    private JLocalVariable localVar;
}
