/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import at.dms.classfile.NoArgInstruction;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.JavaStyleComment;

/**
 * JLS 14.16: Throw Statement.
 *
 * A throw statement causes an exception to be thrown.
 */
public class JThrowStatement
    extends JStatement
{
    /**
     * ...
     */
    private JExpression expression;

    /**
     * Construct a node in the parsing tree.
     *
     * @param where
     * The line of this node in the source code.
     *
     * @param expression 
     * The expression to throw.
     *
     * @param comments
     * Java implementation comments for this statement.
     */
    public JThrowStatement(TokenReference where,
                           JExpression expression,
                           JavaStyleComment[] comments)
    {
        super(where, comments);
        this.expression = expression;
    }

    /**
     * Check statement.
     *
     * @param context
     * The actual context of the check.
     *
     * @exception PositionedError
     * If the check fails.
     */
    public void checkStatement(CBodyContext context)
        throws PositionedError
    {
        expression =
            expression.checkExpression(context.createExpressionContext());

        /* Throw argument must be a reference type that is a subclass of
         * java.lang.Throwable.
         */
        check(context,
              expression.getType().isReference()
              && expression.isAssignableTo(CStdType.Throwable),
              KjcMessages.THROW_BADTYPE, expression.getType());

        /* If expression is not Error or RuntimeException, it is a "checked"
         * exception that must be declared in method declarations if it isn't
         * caught.
         */
        if (expression.getType().isCheckedException()) {
            context.addThrowable
                (new CThrowableInfo((CClassType)expression.getType(),
                                    this, context.makeSnapshot()));
        }
        context.setReachable(false);
    }

    /**
     * Accepts the specified visitor.
     *
     * @param p
     * The visitor.
     */
    public void accept(KjcVisitor p)
    {
        super.accept(p);
        p.visitThrowStatement(this, expression);
    }

    /**
     * Generates a sequence of bytecodes.
     *
     * @param code
     * The code list.
     */
    public void genCode(CodeSequence code)
    {
        code.setLineNumber(getTokenReference().getLine());

        expression.genCode(code);
        code.plantInstruction(new NoArgInstruction(opc_athrow));
    }
}
