/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.lexgen;

import java.io.BufferedInputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.InputStream;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.HashMap;
import java.util.Map;
import java.util.List;

/**
 * A token definition file for the lexer generator.
 */
class DefinitionFile
{
    /**
     *
     */
    private static final String DEFAULT_PREFIX = "LITERAL_";

    /**
     *
     */
    private final String sourceFile;

    /**
     *
     */
    private final String packageName;

    /**
     *
     */
    private final String vocabulary;

    /**
     *
     */
    private final String prefix;

    /**
     *
     */
    private final TokenDefinition[] definitions;

    /**
     * Constructs a token definition file.
     */
    public DefinitionFile(String sourceFile,
                          String packageName,
                          String vocabulary,
                          String prefix,
                          List definitions)
    {
        this.sourceFile = sourceFile;
        this.packageName = packageName;
        this.vocabulary = vocabulary;
        this.prefix = (prefix == null ? DEFAULT_PREFIX : prefix);

        this.definitions = (TokenDefinition[])
            definitions.toArray(new TokenDefinition[definitions.size()]);
    }

    /**
     * Reads and parses a token definition file.
     *
     * @param sourceFile
     * The name of the source file.
     * @return
     * A class info structure holding the information from the source.
     */
    public static DefinitionFile read(String sourceFile)
        throws LexgenError
    {
        try {
            InputStream input = new BufferedInputStream
                (new FileInputStream(sourceFile));
            LexgenLexer scanner = new LexgenLexer(input);
            LexgenParser parser = new LexgenParser(scanner);
            DefinitionFile defs = parser.aCompilationUnit(sourceFile);

            input.close();

            return defs;
        }
        catch (FileNotFoundException e) {
            throw new LexgenError(LexgenMessages.FILE_NOT_FOUND, sourceFile);
        }
        catch (IOException e) {
            throw new LexgenError(LexgenMessages.IO_EXCEPTION, sourceFile,
                                  e.getMessage());
        }
        catch (antlr.ParserException e) {
            throw new LexgenError(LexgenMessages.PARSING_ERROR,
                                  sourceFile,
                                  e.getLine() + " : " + e.getErrorMessage());
        }

    }

    /**
     * Check for duplicate identifiers.
     *
     * @param identifiers
     * A table of all token identifiers.
     *
     * @param prefix
     * The literal prefix.
     *
     * @param id
     * The id of the first token.
     *
     * @return
     * The id of the last token + 1
     */
    public int checkIdentifiers(Map identifiers, String prefix, int id)
        throws LexgenError
    {
        for (int i = 0; i < definitions.length; i++) {
            definitions[i].checkIdentifiers(identifiers, prefix, id++,
                                            sourceFile);
        }

        return id;
    }

    /**
     * Prints token definitions to definition file.
     *
     * @param out the output stream.
     * @param prefix the literal prefix.
     */
    public void printDefinition(PrintWriter out, String prefix)
    {
        out.println("/* Definitions from " + sourceFile);
        out.println(" */");

        for (int i = 0; i < definitions.length; i++) {
            definitions[i].printDefinition(out, prefix);
        }
    }

    /**
     * Prints the token definition to interface file.
     *
     * @param out The output stream.
     * @param parent The super interface.
     */
    public void printInterface(PrintWriter out, String parent)
    {
        out.println("/* Generated from " + sourceFile);
        out.println(" */");
        out.println();
        out.println("package " + packageName + ";");
        out.println();
        out.println("public interface " + vocabulary + "TokenTypes");
        if (parent != null) {
            out.println("    extends " + parent);
        }
        out.println("{");

        if (parent == null) {
            out.println("    int EOF = 1;");
            out.println("    int NULL_TREE_LOOKAHEAD = 3;");
        }

        for (int i = 0; i < definitions.length; i++) {
            definitions[i].printInterface(out, this.prefix);
        }

        out.println("}");
    }

    /**
     * Adds keywords to list.
     */
    public void putKeywords(List keywords, List types, String prefix)
    {
        for (int i = 0; i < definitions.length; i++) {
            definitions[i].putKeyword(keywords, types, prefix);
        }
    }

    /**
     * Returns the class name.
     */
    public String getClassName()
    {
        return packageName + "." + vocabulary + "TokenTypes";
    }

    /**
     * Returns the package name.
     */
    public String getPackageName()
    {
        return packageName;
    }

    /**
     * Returns the vocabulary name,
     */
    public String getVocabulary()
    {
        return vocabulary;
    }

    /**
     * Returns the literal prefix.
     */
    public String getPrefix()
    {
        return prefix;
    }
}
