/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.lexgen;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.HashMap;
import java.util.Map;
import java.util.ArrayList;
import java.util.List;

import at.dms.jperf.JPerf;

/**
 * This class is the entry point for the KOPI lexer generator.
 */
public class Main
{
    /**
     * Command-line options.
     */
    private LexgenOptions  options;

    /**
     * Parsed definitions.
     */
    private DefinitionFile[] definitions;

    private static final int MIN_TOKEN_ID = 4;
    /**
     * Only main can construct Main
     */
    private Main(String[] args)
    {
        if (!parseArguments(args)) {
            System.exit(1);
        }

        if (! options.definition && !options.inter && !options.keywords) {
            options.usage();
            System.exit(1);
        }

        String[] infiles = options.nonOptions;

        if (infiles.length == 0) {
            options.usage();
            System.err.println(LexgenMessages.NO_SOURCE_FILE.getFormat());
            System.exit(1);
        } else {
            parseSourceFiles(infiles);
            checkIdentifiers();
            if (options.definition) {
                buildDefinitionFile();
            }
            if (options.inter) {
                buildInterfaceFile();
            }
            if (options.keywords) {
                buildKeywordFile();
            }
            System.exit(0);
        }
    }

    /**
     * Parsed DefinitionFile objects from source.
     */
    private void parseSourceFiles(String[] infiles)
    {
        boolean  errorsFound = false;

        definitions = new DefinitionFile[infiles.length];

        for (int i = 0; i < infiles.length; i++) {
            try {
                definitions[i] = DefinitionFile.read(infiles[i]);
            }
            catch (LexgenError e) {
                System.err.println(e.getMessage());
                errorsFound = true;
            }
        }

        if (errorsFound) {
            System.exit(1);
        }
    }


    /**
     * Checks identifiers from parsed source.
     */
    private void checkIdentifiers()
    {
        Map identifiers = new HashMap();

        boolean errorsFound = false;
        String  prefix = definitions[definitions.length - 1].getPrefix();
        int   cnt = MIN_TOKEN_ID;

        for (int i = 0; i < definitions.length; i++) {
            try {
                cnt = definitions[i].checkIdentifiers(identifiers,
                                                      prefix, cnt);
            }
            catch (LexgenError e) {
                System.err.println(e.getMessage());
                errorsFound = true;
            }
        }

        if (errorsFound) {
            System.exit(1);
        }
    }

    /**
     * Build output file TokenTypes.txt
     */
    private void buildDefinitionFile()
    {
        DefinitionFile definition = definitions[definitions.length - 1];
        String  vocabulary = definition.getVocabulary();
        String  prefix = definition.getPrefix();

        File  outputFile = new File(vocabulary + "TokenTypes.txt");

        try {
            PrintWriter out = new PrintWriter(new FileWriter(outputFile));

            // !!! print file header

            // print vocabulary name
            out.println(vocabulary + " // output vocabulary");

            for (int i = 0; i < definitions.length; i++) {
                definitions[i].printDefinition(out, prefix);
            }

            out.flush();
            out.close();
        }
        catch (IOException e) {
            System.err.println("I/O Exception on " + outputFile.getPath()
                               + ": " + e.getMessage());
            System.exit(1);
        }
    }

    /**
     * Build output file TokenTypes.java
     */
    private void buildInterfaceFile()
    {
        DefinitionFile definition = definitions[definitions.length - 1];
        String vocabulary = definition.getVocabulary();
        File outputFile = new File(vocabulary + "TokenTypes.java");

        try {
            PrintWriter out = new PrintWriter(new FileWriter(outputFile));

            // !!! print file header

            definition.printInterface(out,
                                      ( definitions.length == 1
                                        ? null
                                        : definitions[definitions.length - 2]
                                        .getClassName() ));

            out.flush();
            out.close();
        }
        catch (IOException e) {
            System.err.println("I/O Exception on " + outputFile.getPath()
                               + ": " + e.getMessage());
            System.exit(1);
        }
    }

    /**
     *
     */
    private void buildKeywordFile()
    {
        DefinitionFile definition = definitions[definitions.length - 1];
        String vocabulary = definition.getVocabulary();
        String prefix = definition.getPrefix();
        String packageName = definition.getPackageName();

        List keywords = new ArrayList();
        List types = new ArrayList();
        List header = new ArrayList();
        List footer = new ArrayList();

        // retrieve keywords to hash
        for (int i = 0; i < definitions.length; i++) {
            definitions[i].putKeywords(keywords, types, prefix);
        }

        /* header */
        header.add("/* Generated by LexGen/JPerf");
        header.add("*/");
        header.add("package " + packageName + ";");
        header.add("import at.dms.compiler.CToken;\n");
        header.add("final class " + vocabulary + "Keywords implements "
                   + vocabulary + "TokenTypes");
        header.add("{");
        header.add("    public static CToken lookup(final char[] data, "
                   +                                "int offset, int length)");
        header.add("    {");
        header.add("        return tokens[find(data, offset, length) + 1];");
        header.add("    }");
        header.add("    private static final CToken[] tokens = new CToken[]");
        header.add("    {");
        header.add("        null,");

        for (int i = 0; i < types.size(); i++) {
            String line = "        new CToken(" + (String)types.get(i)
                + ", \"" + (String)keywords.get(i) + "\")";

            header.add(line + ",");
        }
        header.add("    };");

        // footer
        footer.add("}");

        // generate perfect hash function
        JPerf perf = new JPerf(makeStringArray(keywords),
                               makeStringArray(header),
                               makeStringArray(footer));
        perf.build();

        try {
            perf.genCode(vocabulary + "Keywords.java");
        }
        catch (IOException e) {
            System.err.println("I/O Exception on " + vocabulary
                               + "Keywords.java" + ": " + e.getMessage());
            System.exit(1);
        }
    }

    /**
     * Transforms a List of strings into a string array.
     */
    private static String[] makeStringArray(List list)
    {
        return (String[])list.toArray(new String[list.size()]);
    }

    /**
     * Entry point to the lexer generator.
     */
    public static void main(String[] args)
    {
        new Main(args);
    }

    /**
     * Parse command line arguments.
     */
    private boolean parseArguments(String[] args)
    {
        options = new LexgenOptions();
        return options.parseCommandLine(args);
    }
}
