/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.dis;

import java.util.ArrayList;
import java.util.List;

import at.dms.classfile.AccessorContainer;
import at.dms.classfile.AbstractInstructionAccessor;
import at.dms.classfile.ClassConstant;
import at.dms.classfile.ClassRefInstruction;
import at.dms.classfile.FieldRefConstant;
import at.dms.classfile.FieldRefInstruction;
import at.dms.classfile.IincInstruction;
import at.dms.classfile.Instruction;
import at.dms.classfile.InterfaceConstant;
import at.dms.classfile.InvokeinterfaceInstruction;
import at.dms.classfile.JumpInstruction;
import at.dms.classfile.LineNumberInfo;
import at.dms.classfile.LocalVarInstruction;
import at.dms.classfile.MethodRefConstant;
import at.dms.classfile.MethodRefInstruction;
import at.dms.classfile.MultiarrayInstruction;
import at.dms.classfile.NewarrayInstruction;
import at.dms.classfile.NoArgInstruction;
import at.dms.classfile.PushLiteralInstruction;
import at.dms.classfile.SwitchInstruction;

/**
 * 
 */
class InstructionHandle
    extends AbstractInstructionAccessor
{
    /**
     * ...
     */
    private final Instruction instruction;

    /**
     * ...
     */
    private final int address;

    /**
     * ...
     */
    private List lineNumbers;

    /**
     * ...
     */
    private boolean isTarget;

    /**
     * Creates a new instruction handle.
     *
     * @param instruction
     * The instruction.
     * @param address
     * The address of the instruction.
     */
    public InstructionHandle(Instruction instruction, int address)
    {
        this.instruction = instruction;
        this.address = address;

        this.isTarget = false;
    }

    /**
     * Notifies this handle that is has been attached to the specified
     * container.
     */
    public void attachTo(AccessorContainer container)
    {
        if (container instanceof LineNumberInfo) {
            if (lineNumbers == null) {
                lineNumbers = new ArrayList();
            }
            lineNumbers.add
                (new Integer(((LineNumberInfo)container).getLine()));
        }
        else {
            // Declares this instruction to be target of a jump, handler, ...
            this.isTarget = true;
        }
    }

    /**
     * Returns an identifier for the instruction address.
     */
    public String getLabel()
    {
        return "_L" + address;
    }

    /**
     * Prints a byte code instruction.
     */
    public void write(IndentingWriter out, boolean optionShowStack)
    {
        if (lineNumbers != null) {
            for (int i = 0; i < lineNumbers.size(); i++) {
                out.println();
                out.print("@line " + (Integer)lineNumbers.get(i));
            }
        }

        out.println();
        if (isTarget) {
            out.print(getLabel() + ":");
        }
        if (optionShowStack) {
            out.print(" /* " + "N/A" + " */ ");
        }
        out.print(" ");
        writeInstruction(out, instruction);
    }

    /**
     * Prints an instruction (dispatches to instruction types.)
     */
    private void writeInstruction(IndentingWriter out,
                                  Instruction instruction)
    {
        if (instruction instanceof ClassRefInstruction) {
            writeClassRefInstruction(out, (ClassRefInstruction)instruction);
        }
        else if (instruction instanceof FieldRefInstruction) {
            writeFieldRefInstruction(out, (FieldRefInstruction)instruction);
        }
        else if (instruction instanceof IincInstruction) {
            writeIincInstruction(out, (IincInstruction)instruction);
        }
        else if (instruction instanceof InvokeinterfaceInstruction) {
            writeInvokeinterfaceInstruction
                (out, (InvokeinterfaceInstruction)instruction);
        }
        else if (instruction instanceof JumpInstruction) {
            writeJumpInstruction(out, (JumpInstruction)instruction);
        }
        else if (instruction instanceof LocalVarInstruction) {
            writeLocalVarInstruction(out, (LocalVarInstruction)instruction);
        }
        else if (instruction instanceof MethodRefInstruction) {
            writeMethodRefInstruction(out, (MethodRefInstruction)instruction);
        }
        else if (instruction instanceof MultiarrayInstruction) {
            writeMultiarrayInstruction(out,
                                       (MultiarrayInstruction)instruction);
        }
        else if (instruction instanceof NewarrayInstruction) {
            writeNewarrayInstruction(out, (NewarrayInstruction)instruction);
        }
        else if (instruction instanceof NoArgInstruction) {
            writeNoArgInstruction(out, (NoArgInstruction)instruction);
        }
        else if (instruction instanceof PushLiteralInstruction) {
            writePushLiteralInstruction(out,
                                        (PushLiteralInstruction)instruction);
        }
        else if (instruction instanceof SwitchInstruction) {
            writeSwitchInstruction(out, (SwitchInstruction)instruction);
        }
        else {
            writeOpcode(out, "@unknown");
            out.print("opcode=" + instruction.getOpcode());
        }
    }

    /**
     * Prints the opcode.
     */
    private void writeOpcode(IndentingWriter out, String opcode)
    {
        out.print(opcode);
        out.print(opcode.length() < 8 ? "  " : " ");
    }

    /**
     * Prints the opcode.
     */
    private void writeOpcode(IndentingWriter out, int opcode)
    {
        writeOpcode(out, OpcodeNames.getName(opcode));
    }

    /**
     * Prints a ClassRefInstruction.
     */
    private void writeClassRefInstruction(IndentingWriter out,
                                          ClassRefInstruction instruction)
    {
        writeOpcode(out, instruction.getOpcode());
        printClassRef(out, instruction.getClassConstant());
    }

    /**
     * Prints a FieldRefInstruction.
     */
    private void writeFieldRefInstruction(IndentingWriter out,
                                          FieldRefInstruction instruction)
    {
        writeOpcode(out, instruction.getOpcode());
        printFieldRef(out, instruction.getFieldRefConstant());
    }

    /**
     * Prints a IincInstruction.
     */
    private void writeIincInstruction(IndentingWriter out,
                                      IincInstruction instruction)
    {
        writeOpcode(out, instruction.getOpcode());
        out.print(instruction.getVariable());
        out.print(" ");
        out.print(instruction.getIncrement());
    }

    /**
     * Prints a InvokeinterfaceInstruction.
     */
    private void writeInvokeinterfaceInstruction(IndentingWriter out,
                                                 InvokeinterfaceInstruction in)
    {
        writeOpcode(out, in.getOpcode());
        printInterfaceRef(out, in.getInterfaceConstant());
        out.print(" ");
        out.print(in.getNbArgs());
    }

    /**
     * Prints a JumpInstruction.
     */
    private void writeJumpInstruction(IndentingWriter out,
                                      JumpInstruction instruction)
    {
        writeOpcode(out, instruction.getOpcode());
        printInstructionHandle(out,
                               (InstructionHandle)instruction.getTarget());
    }

    /**
     * Prints a LocalVarInstruction.
     */
    private void writeLocalVarInstruction(IndentingWriter out,
                                          LocalVarInstruction instruction)
    {
        writeOpcode(out, instruction.getOpcode());
        out.print(instruction.getIndex());
    }

    /**
     * Prints a MethodRefInstruction.
     */
    private void writeMethodRefInstruction(IndentingWriter out,
                                           MethodRefInstruction instruction)
    {
        writeOpcode(out, instruction.getOpcode());
        printMethodRef(out, instruction.getMethodRefConstant());
    }

    /**
     * Prints a MultiarrayInstruction.
     */
    private void writeMultiarrayInstruction(IndentingWriter out,
                                            MultiarrayInstruction instruction)
    {
        writeOpcode(out, instruction.getOpcode());
        printFieldSignature(out, instruction.getType());
        out.print(" ");
        out.print(instruction.getDimension());
    }

    /**
     * Prints a NewarrayInstruction.
     */
    private void writeNewarrayInstruction(IndentingWriter out,
                                          NewarrayInstruction instruction)
    {
        writeOpcode(out, instruction.getOpcode());
        out.print(instruction.getType());
    }

    /**
     * Prints a NoArgInstruction.
     */
    private void writeNoArgInstruction(IndentingWriter out,
                                       NoArgInstruction instruction)
    {
        writeOpcode(out, instruction.getOpcode());
    }

    /**
     * Prints a PushLiteralInstruction.
     */
    private void writePushLiteralInstruction(IndentingWriter out,
                                             PushLiteralInstruction instruct)
    {
        writeOpcode(out, instruct.getOpcode());
        out.print(Disassembler.convertLiteral(instruct.getLiteral()));
    }

    /**
     * Prints a SwitchInstruction.
     */
    private void writeSwitchInstruction(IndentingWriter out,
                                        SwitchInstruction instruction)
    {
        writeOpcode(out, instruction.getOpcode());
        for (int i = 0; i < instruction.getSwitchCount(); i++) {
            out.println();
            out.print("   " + instruction.getMatch(i) + ": ");
            printInstructionHandle
                (out, (InstructionHandle)instruction.getTarget(i));
        }
        out.println();
        out.print("   @default: ");
        printInstructionHandle(out,
                               (InstructionHandle)instruction.getTarget(-1));
    }

    /**
     * Prints a ClassConstant.
     */
    private void printClassRef(IndentingWriter out, ClassConstant ref)
    {
        String  name = ref.getName();

        if (name.charAt(0) == '[') {
            printFieldSignature(out, name);
        }
        else {
            printQualifiedName(out, name);
        }
    }

    private void printFieldRef(IndentingWriter out, FieldRefConstant ref)
    {
        printFieldSignature(out, ref.getType());
        out.print(" ");
        printQualifiedName(out, ref.getName());
    }

    private void printInterfaceRef(IndentingWriter out, InterfaceConstant ref)
    {
        printMethodSpec(out, ref.getName(), ref.getType());
    }

    private void printMethodRef(IndentingWriter out, MethodRefConstant ref)
    {
        printMethodSpec(out, ref.getName(), ref.getType());
    }

    private void printMethodSpec(IndentingWriter out,
                                 String name, String signature)
    {
        String[] type = Disassembler.convertMethodSignature(signature);

        out.print(type[1] + " " + Disassembler.convertQualifiedName(name)
                  + type[0]);
    }

    private void printFieldSignature(IndentingWriter out, String signature)
    {
        out.print(Disassembler.convertFieldSignature(signature));
    }

    private void printQualifiedName(IndentingWriter out, String ident)
    {
        out.print(Disassembler.convertQualifiedName(ident));
    }

    private void printInstructionHandle(IndentingWriter out,
                                        InstructionHandle handle)
    {
        out.print(handle.getLabel());
    }
}
