/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.msggen;

import java.io.BufferedInputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.InputStream;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;
import java.util.List;

/**
 * A parsed representation of a file containing compiler messages.
 */
class DefinitionFile
{
    /**
     * The source of this definition object.
     */
    private final String sourceFile;

    /**
     * The package name for the resulting java source file.
     */
    private final String packageName;

    /**
     * The prefix for the classname of the resulting java source file.
     */
    private final String prefix;

    /**
     * The messages that were read from the source file.
     */
    private final MessageDefinition[] definitions;

    /**
     * Constructs a token definition file
     */
    public DefinitionFile(String sourceFile,
                          String packageName,
                          String prefix,
                          List definitions)
    {
        this.sourceFile = sourceFile;
        this.packageName = packageName;
        this.prefix = (prefix == null ? "" : prefix);

        this.definitions = (MessageDefinition[])
            definitions.toArray(new MessageDefinition[definitions.size()]);
    }

    /**
     * Reads and parses an token definition file
     *
     * @param sourceFile the name of the source file
     * @return a class info structure holding the information from the source
     *
     */
    public static DefinitionFile read(String sourceFile)
        throws MsggenError
    {
        try {
            InputStream input = new BufferedInputStream
                (new FileInputStream(sourceFile));
            MsggenLexer scanner = new MsggenLexer(input);
            MsggenParser parser = new MsggenParser(scanner);
            DefinitionFile defs = parser.aCompilationUnit(sourceFile);

            input.close();

            return defs;
        }
        catch (FileNotFoundException e) {
            throw new MsggenError(MsggenMessages.UNKNOWN_FILE, sourceFile);
        }
        catch (IOException e) {
            throw new MsggenError(MsggenMessages.IO_EXCEPTION,
                                  sourceFile, e.getMessage());
        }
        catch (antlr.ParserException e) {
            throw new MsggenError(MsggenMessages.PARSING_ERROR,
                                  sourceFile,
                                  e.getLine() + " : " + e.getErrorMessage());
        }

    }

    /**
     * Check for duplicate identifiers
     * @param identifiers a table of all token identifiers
     * @return the id of the last token + 1
     */
    public void checkIdentifiers(Map identifiers)
        throws MsggenError
    {
        for (int i = 0; i < definitions.length; i++) {
            definitions[i].checkIdentifiers(identifiers, sourceFile);
        }
    }

    /**
     * Prints the token definition to interface file (java)
     * @param out     the output stream
     * @param parent      the super interface
     */
    public void printFile(PrintWriter out)
    {
        out.println("/* Generated from source file " + sourceFile);
        out.println(" * For copyright and licensing information, see the "
                    + "source file.");
        out.println(" */");
        out.println("package " + packageName + ";");
        out.println();
        out.println("import at.dms.util.MessageDescription;");
        out.println();
        out.println("/**");
        out.println(" * Generated file containing message definitions.");
        out.println(" * Source file: <code>" + sourceFile + "</code>");
        out.println(" */");
        out.println("public class " + prefix + "Messages");
        out.println("{");

        for (int i = 0; i < definitions.length; i++) {
            definitions[i].printInterface(out, this.prefix);
        }

        out.println("}");
    }

    /**
     * Returns the package name
     */
    public String getClassName()
    {
        return packageName + "." + prefix + "Messages";
    }

    /**
     * Returns the package name
     */
    public String getPackageName()
    {
        return packageName;
    }

    /**
     * Returns the literal prefix
     */
    public String getPrefix()
    {
        return prefix;
    }

}
