/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.optgen;

import java.io.BufferedInputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.PrintWriter;
import java.util.Map;
import java.util.ArrayList;
import java.util.List;

/**
 * Parsed file containing GNU getopt definitions for a command-line tool.
 */
class DefinitionFile
{
    /**
     * This message is included in the Java source files of option definitions
     * that are generated by this class.
     */
    private final static String HELP_MESSAGE[] = new String[] {
        "        PrintStream err = System.err;",
        "        err.println();",
        "        err.println(\"This program is originally part of the KOPI "
        +                     "suite of Java compilation tools.\");",

        "        err.println(\"For more information on the KOPI suite, "
        +                     "see:\");",
        "        err.println(\"http://www.dms.at/kopi\");" };

    /**
     */
    private final String sourceFile;

    /**
     */
    private final String packageName;

    /**
     */
    private final String parent;

    /**
     */
    private final String prefix;

    /**
     */
    private final String version;

    /**
     */
    private final String usage;

    /**
     */
    private final OptionDefinition[] definitions;

    /**
     * Constructs a definition file.
     */
    public DefinitionFile(String sourceFile,
                          String packageName,
                          String parent,
                          String prefix,
                          String version,
                          String usage,
                          List definitions)
    {
        this.sourceFile = sourceFile;
        this.packageName = packageName;
        this.parent = parent;
        this.version = version;
        this.usage = usage;
        this.prefix = prefix;

        this.definitions = (OptionDefinition[])
            definitions.toArray(new OptionDefinition[definitions.size()]);
    }

    /**
     * Reads and parses an token definition file.
     *
     * @param sourceFile
     * The name of the source file.
     *
     * @return
     * A class info structure holding the information from the source.
     *
     */
    public static DefinitionFile read(String sourceFile)
        throws OptgenError
    {
        try {
            InputStream input = new BufferedInputStream
                (new FileInputStream(sourceFile));
            OptgenLexer scanner = new OptgenLexer(input);
            OptgenParser parser = new OptgenParser(scanner);
            DefinitionFile defs = parser.aCompilationUnit(sourceFile);

            input.close();

            return defs;
        }
        catch (FileNotFoundException e) {
            throw new OptgenError(OptgenMessages.UNKNOWN_FILE, sourceFile);
        }
        catch (IOException e) {
            throw new OptgenError(OptgenMessages.IO_EXCEPTION,
                                  sourceFile, e.getMessage());
        }
        catch (antlr.ParserException e) {
            throw new OptgenError(OptgenMessages.PARSING_ERROR,
                                  sourceFile,
                                  e.getLine() + " : " + e.getErrorMessage());
        }

    }

    /**
     * Check for duplicate identifiers.
     *
     * @param identifiers a table of all token identifiers.
     * @param prefix the literal prefix.
     * @param id the id of the first token.
     */
    public void checkIdentifiers(Map identifiers)
        throws OptgenError
    {
        for (int i = 0; i < definitions.length; i++) {
            definitions[i].checkIdentifiers(identifiers, sourceFile);
        }
    }

    /**
     * Check for duplicate identifiers.
     * @param identifiers a table of all token identifiers.
     * @param prefix the literal prefix
     * @param id the id of the first token
     */
    public void checkShortcuts(Map identifiers)
        throws OptgenError
    {
        for (int i = 0; i < definitions.length; i++) {
            definitions[i].checkShortcuts(identifiers, sourceFile);
        }
    }

    /**
     * Prints the token definition to interface file (java).
     *
     * @param out the output stream.
     */
    public void printFile(PrintWriter out)
    {
        out.println("/* Generated from source file " + sourceFile);
        out.println(" * For copyright and licensing information, see the "
                    + "source file.");
        out.println(" */");
        out.println("package " + packageName + ";");
        out.println();
        out.println("import java.io.PrintStream;");
        out.println();
        out.println("import gnu.getopt.Getopt;");
        out.println("import gnu.getopt.LongOpt;");
        out.println();
        out.println("/**");
        out.println(" * Generated file containing message definitions.");
        out.println(" * Source file: <code>" + sourceFile + "</code>");
        out.println(" */");
        out.println("public class " + prefix + "Options");
        if (parent != null) {
            out.println("    extends " + parent);
        }
        out.println("{");

        /* Print constructors */
        out.println();
        out.println("    public " + prefix + "Options(String name)");
        out.println("    {");
        out.println("        super(name);");
        out.println("    }");
        out.println();
        out.println("    public " + prefix + "Options()");
        out.println("    {");
        out.println("        this(\"" + prefix + "\");");
        out.println("    }");

        for (int i = 0; i < definitions.length; i++) {
            definitions[i].printFields(out);
        }

        /* Process-option method */
        out.println();
        out.println("    public boolean processOption(int code, Getopt g)");
        out.println("    {");
        out.println("        switch (code) {");

        for (int i = 0; i < definitions.length; i++) {
            definitions[i].printParseArgument(out);
        }

        out.println("        default:");
        out.println("            return super.processOption(code, g);");
        out.println("        }");
        out.println("        return true;");
        out.println("    }");


        /* Get-options method */
        out.println();
        out.println("    public String[] getOptions()");
        out.println("    {");
        out.println("        String[] parent = super.getOptions();");
        out.println("        String[] total = new String[parent.length + "
                    +                         definitions.length+ "];");
        out.println("        System.arraycopy(parent, 0, total, 0, "
                    +                         "parent.length);");
        for (int i = 0; i < definitions.length; i++) {
            out.print("        total[parent.length + " + i + "] = ");
            definitions[i].printUsage(out);
            out.println(";");
        }
        out.println();
        out.println("        return total;");
        out.println("    }");

        /* Get-short-options method */
        out.println();
        out.println("    public String getShortOptions()");
        out.println("    {");
        out.print(  "        return \"");
        for (int i = 0; i < definitions.length; i++) {
            definitions[i].printShortOption(out);
        }
        out.println("\" + super.getShortOptions();");
        out.println("    }");

        /* 'version' method */
        if (version != null) {
            out.println();
            out.println("    public void version()");
            out.println("    {");
            out.println("        System.out.println(" + version + ");");
            out.println("    }");
        }

        /* 'usage' method */
        out.println();
        out.println("    public void usage()");
        out.println("    {");
        out.println("        System.err.println(" + usage + ");");
        out.println("        version();");
        out.println("    }");

        /* 'help' method */
        out.println();
        out.println("    public void help()");
        out.println("    {");
        out.println("        System.err.println(" + usage + ");");
        out.println("        printOptions();");
        for (int i = 0; i < HELP_MESSAGE.length; i++) {
            out.println(HELP_MESSAGE[i]);
        }
        out.println("        version();");
        out.println("  }");

        /* Get-long-options method */
        out.println();
        out.println("    public LongOpt[] getLongOptions()");
        out.println("    {");
        out.println("        LongOpt[] parent = super.getLongOptions();");
        out.println("        LongOpt[] total = new LongOpt[parent.length + "
                    +                          "LONGOPTS.length];");
        out.println();
        out.println("        System.arraycopy(parent, 0, total, 0, "
                    +                        "parent.length);");
        out.println("        System.arraycopy(LONGOPTS, 0, total, "
                    +                      "parent.length, LONGOPTS.length);");
        out.println();
        out.println("        return total;");
        out.println("    }");

        /* LONGTOPS field declaration */
        out.println();
        out.println("    private static final LongOpt[] LONGOPTS = {");
        for (int i = 0; i < definitions.length; i++) {
            if (i != 0) {
                out.println(",");
            }
            definitions[i].printLongOpts(out);
        }
        out.println();
        out.println("    };");

        out.println("}");
    }

    /**
     * Returns the class name.
     */
    public String getClassName()
    {
        return packageName + "." + prefix + "Options";
    }

    /**
     * Returns the package name.
     */
    public String getPackageName()
    {
        return packageName;
    }

    /**
     * Returns the literal prefix.
     */
    public String getPrefix()
    {
        return prefix;
    }

}
