/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.optgen;

import java.io.IOException;
import java.io.PrintWriter;
import java.util.Map;

/**
 * Declaration of one option, parsed from option definition file.
 */
class OptionDefinition
{
    /**
     *
     */
    private final String longname;

    /**
     *
     */
    private final String shortname;

    /**
     *
     */
    private final String type;

    /**
     *
     */
    private final String defaultValue;

    /**
     *
     */
    private final String argument;

    /**
     *
     */
    private final String help;

    /**
     * Constructs an option definition.
     */
    public OptionDefinition(String longname,
                            String shortname,
                            String type,
                            String defaultValue,
                            String argument,
                            String help)
    {
        this.longname = trail(longname);
        this.shortname = trail(shortname);
        this.type = type;
        this.defaultValue = trail(defaultValue);
        this.argument = trail(argument);
        this.help = trail(help);
    }

    /**
     * Strip leading and trailing quotes.
     */
    private static final String trail(String s)
    {
        if (s == null) {
            return null;
        }
        else if (s.length() < 2) {
            return s;
        }
        else {
            return s.substring(1, s.length() - 1);
        }
    }

    /**
     * Check for duplicate identifiers.
     *
     * @param identifiers a table of all token identifiers
     * @param sourceFile  the file where the token is defined
     */
    public void checkIdentifiers(Map identifiers, String sourceFile)
        throws OptgenError
    {
        String stored = (String)identifiers.get(longname);

        if (stored != null) {
            throw new OptgenError(OptgenMessages.DUPLICATE_DEFINITION,
                                  new Object[] { longname, sourceFile,
                                                 stored });
        }

        identifiers.put(longname, sourceFile);
    }

    /**
     * Check for duplicate identifiers.
     *
     * @param identifiers a table of all token identifiers.
     * @param sourceFile  the file where the token is defined.
     */
    public void checkShortcuts(Map shortcuts, String sourceFile)
        throws OptgenError
    {
        String      stored = (String)shortcuts.get(shortname);

        if (stored != null) {
            throw new OptgenError(OptgenMessages.DUPLICATE_SHORTCUT,
                                  new Object[] { shortname, sourceFile,
                                                 stored });
        }
        shortcuts.put(shortname, sourceFile);
    }

    /**
     * Prints the token definition to interface file (java).
     *
     * @param out
     * The output stream.
     */
    public void printParseArgument(PrintWriter out)
    {
        out.println("        case \'" + shortname + "\':");
    
        out.println("            " + longname + " =");
        if (argument == null) {
            out.println("                !" + defaultValue + ";");
        }
        else {
            String methodName;
            String argument = this.argument;

            if (type.equals("int")) {
                methodName = "getInt";
                if (argument == null || argument.equals("")) {
                    argument = "0";
                }
            }
            else {
                methodName = "getString";
                argument = "\"" + argument + "\"";
            }

            out.println("                " + methodName + "(g, " + argument
                        + ");");
        }
        out.println("            break;");
    }

    /**
     * Prints the token definition to interface file (java).
     *
     * @param out the output stream.
     */
    public void printFields(PrintWriter out)
    {
        out.print("    public ");
        out.print(type);
        out.print(" ");
        out.print(longname);
        out.print(" = ");
        if (!type.equals("String") || defaultValue.equals("null")) {
            out.print(defaultValue);
        }
        else {
            out.print("\"" + defaultValue + "\"");      
        }
        out.println(";");
    }

    /**
     * Prints the token definition to interface file (java).
     *
     * @param out the output stream.
     */
    public void printUsage(PrintWriter out)
    {
        String prefix = "\"  --";

        prefix += longname;
        prefix += ", -";
        prefix += shortname;
        if (argument != null) {
            prefix += "<" + type + ">";
        }
        prefix += ": ";
        for (int i = prefix.length(); i < 25; i++) {
            prefix += " ";
        }
        out.print(prefix + help);
        if (!defaultValue.equals("null")) {
            out.print(" [");
            out.print(defaultValue);
            out.print("]");
        }
        out.print("\"");
    }

    public void printLongOpts(PrintWriter out)
    {
        out.print("        new LongOpt(\"");
        out.print(longname);
        out.print("\", ");
        if (argument == null) {
            out.print("LongOpt.NO_ARGUMENT");
        }
        else if (argument.equals("")) {
            out.print("LongOpt.REQUIRED_ARGUMENT");
        }
        else {
            out.print("LongOpt.OPTIONAL_ARGUMENT");
        }
        out.print(", null, \'");
        out.print(shortname);    
        out.print("\')");
    }

    public void printShortOption(PrintWriter out)
    {
        out.print(shortname);
        if (argument != null) {
            if (argument.equals("")) {
                out.print(":");
            }
            else {
                out.print("::");
            }
        }
    }
}
